
# Dash app to investigate graphs


def graph(G, mode="external", **kwargs):
    """
    G: a multidirectional graph

    kwargs are passed to the Jupyter_Dash.run_server() function. Some usefull arguments are:
        mode: "inline" to run app inside the jupyter nodebook, default is external 
        debug: True or False, Usefull to catch errors during development.
    """

    import dash
    from jupyter_dash import JupyterDash
    import dash_cytoscape as cyto
    from dash.dependencies import Output, Input
    import dash_html_components as html
    import dash_core_components as dcc
    import dash_table
    import networkx as nx
    import scConnect as cn
    import plotly.graph_objs as go
    import plotly.io as pio
    import pandas as pd
    import numpy as np
    import json
    import matplotlib
    import matplotlib.pyplot as plt
    pio.templates.default = "plotly_white"

    cyto.load_extra_layouts()

    JupyterDash.infer_jupyter_proxy_config()

    app = JupyterDash(__name__)

    server = app.server
    # Add a modified index string to change the title to scConnect
    app.index_string = '''
        <!DOCTYPE html>
        <html>
            <head>
                {%metas%}
                <title>scConnect</title>
                {%favicon%}
                {%css%}
            </head>
            <body>
                {%app_entry%}
                <footer>
                    {%config%}
                    {%scripts%}
                    {%renderer%}
            </body>
        </html>
        '''
    # Add colors to each node
    nodes = pd.Categorical(G.nodes())
    # make a list of RGBA tuples, one for each node
    colors = plt.cm.tab20c(nodes.codes/len(nodes.codes), bytes=True)
    # zip node to color
    color_map_nodes = dict(zip(nodes, colors))

    # add these colors to original graph
    for node, color in color_map_nodes.items():
        G.nodes[node]["color"] = color[0:3]  # Save only RGB

    # Add colors to edges(source node color) for  G
    for u, v, k in G.edges(keys=True):
        G.edges[u, v, k]["color"] = color_map_nodes[u][0:3]

    # load graph into used formes
    def G_to_flat(G, weight):
        G_flat = cn.graph.flatten_graph(G, weight=weight, log=True)

        # Add colors to edges(source node color) for G_flat
        for u, v, in G_flat.edges():
            G_flat.edges[u, v]["color"] = color_map_nodes[u][0:3]
        return G_flat

    # produce full graph variante to extract metadata
    G_flat =G_to_flat(G, weight="score")
    G_split = cn.graph.split_graph(G)

    # find and sort all found interactions
    interactions = list(G_split.keys())
    interactions.sort()

    G_cyto = nx.cytoscape_data(G_flat)

    # get min and max weight for all edges for flat and normal graph
    #weights = [d["weight"] for u, v, d in G_flat.edges(data=True)]
    scores = [d["score"] for u, v, d in G.edges(
        data=True)]
    cent = [d["centrality"] for n, d in G.nodes(data=True)]

    # prepare data for network graph
    nodes = G_cyto["elements"]["nodes"]
    elements = []

    # collect all available genes
    genes = list(nodes[0]["data"]["genes"].keys())

    # Styling parameters
    font_size = 20

    # Style for network graph
    default_stylesheet = [
        {
            'selector': 'node',
            'style': {
                'background-color': 'data(color)',
                'label': 'data(id)',
                'shape': 'ellipse',
                'opacity': 1,
                'font-size': f'{font_size}',
                'font-weight': 'bold',
                'text-wrap': 'wrap',
                'text-max-width': "100px",
                'text-opacity': 1,
                'text-outline-color': "white",
                'text-outline-opacity': 1,
                'text-outline-width': 2
            }
        },

        {
            'selector': 'node:selected',
            'style': {
                'background-color': 'data(color)',
                'label': 'data(id)',
                'shape': 'ellipse',
                'opacity': 1,
                'border-color': "black",
                'border-width': "5"
            }
        },

        {
            'selector': 'edge',
            'style': {
                'line-color': 'data(color)',
                "opacity": 0.7,
                "curve-style": "unbundled-bezier",
                "width": "data(weight)",
                "target-arrow-shape": "vee",
                "target-arrow-color": "black",
                'z-index': 1,
                'font-size': f'{font_size}'
            }
        },

        {
            'selector': 'edge:selected',
            'style': {
                'line-color': 'red',
                'line-style': "dashed",
                'opacity': 1,
                'z-index': 10,
            }
        }
    ]
    app.layout = html.Div(className="wrapper", children=[  # wrapper

        html.Div(className="header", children=[  # header
            html.Img(src="assets/logo.png", alt="scConnect logo"),

            html.Div(className="graph-info", id="graph-stat", children=[
                html.H3(
                    f'Loaded graph with {len(G.nodes())} nodes and {len(G.edges())} edges')
            ])
        ]),

        html.Div(className="network-settings", children=[  # network settings
            html.H2("Network settings", style={"text-align": "center"}),

            html.Label("Interactions"),
            dcc.Dropdown(
                id="network-interaction",
                options=[{'label': "all interactions", 'value': "all"}] +
                [{'label': interaction, 'value': interaction}
                    for interaction in interactions],
                value="all"
            ),
            # select if only significant ligands and receptors should be shown
            html.Label("Graph weight:"),
            dcc.RadioItems(id="weight-select", 
                options=[
                    {"label": "Score", "value": "score"},
                    {"label": "Log score", "value": "log_score"},
                    {"label": "Specificity", "value": "specificity"},
                    {"label": "Importance", "value": "importance"}],
                value="importance",
                labelStyle={
                    'display': 'block',
                    "margin-left": "50px"
                },
                style={"padding": "10px", "margin": "auto"}
            ),

            html.Label("Graph Layout"),
            dcc.Dropdown(
                id="network-layout",
                options=[
                    {'label': name.capitalize(), 'value': name}
                    for name in ['grid', 'random', 'circle', 'cose', 'concentric', 'breadthfirst', 'cose-bilkent', 'cola', 'euler', 'spread', 'dagre', 'klay']
                ],
                value="circle",
                clearable=False),

            html.Label("Weight Filter", style={
                       "paddingBottom": 500, "paddingTop": 500}),
            dcc.Slider(  # min, max and value are set dynamically via a callback
                id="network-filter",
                step=0.001,
                updatemode="drag",
                tooltip={
                    "always_visible": True,
                    "placement": "right"
                },
            ),

            html.Label("Node size"),
            dcc.RangeSlider(
                id="node-size",
                value=[10, 50],
                min=0,
                max=100,
                updatemode="drag"),

            html.Label("Select gene"),
            dcc.Dropdown(
                id="gene_dropdown",
                options=[{"label": gene, "value": gene} for gene in genes],
                clearable=True,
                placeholder="Color by gene expression",
            ),
            
            # Store node colors "hidden" for gene expresison
            html.Div(id="node-colors", style={"display": "none"}, children=[
                ""
            ]),

            html.Div(id="min-max", children=[

            ]),
            # Click to download image of network graph
            html.Button(children="Download current view",  id="download-network-graph", style={"margin":"10px"}) 
            
        ]),  # end network settings
        html.Div(id="network-graph", className="network-graph", children=[  # network graph
            html.H2("Network graph", style={"text-align": "center"}),
            cyto.Cytoscape(
                id="cyto-graph",
                style={
                    'width': '100%',
                    'height': '80vh'},
                stylesheet=default_stylesheet,
                elements=elements,
                autoRefreshLayout=True,
                zoomingEnabled=False)
        ]),  # end network graph

        html.Div(className="sankey-settings", children=[  # network settings
            html.H2("Sankey Settings", style={"text-align": "center"}),

            html.Label("Weight Filter"),
            dcc.Slider(
                id="sankey-filter",
                min=min(scores),
                max=max(scores),
                value=0.75,
                step=0.001,
                updatemode="drag",
                tooltip={
                    "always_visible": True,
                    "placement": "right"
                }),

            html.Label("Toggle weighted"),
            dcc.RadioItems(id="sankey-toggle", options=[
                {"label": "Score", "value": "score"},
                {"label": "Log score", "value": "log_score"},
                {"label": "Specificity", "value": "specificity"},
                {"label": "Importance", "value": "importance"}
            ], value="importance",
            labelStyle={"display": "block"})

        ]),  # end network settings

        html.Div(className="sankey", id="sankey", children=[  # sankey graph
            html.H2("Sankey graph", style={"text-align": "center"}),
            dcc.Graph(id="sankey-graph")
        ]),  # end sankey graph

        html.Div(className="interaction-list", children=[  # interaction list

            html.Div(id="selection", children=[
                html.H2("Interactions", style={"text-align": "center"}),
                html.H3(id="edge-info", style={"text-align": "center"}),
            
                dcc.Graph(id="interaction-scatter"),

                html.Div(id="interaction-selection", style={"display": "none"}, children=[
                ""
                ])
            ]),
            html.Div(children=[
                dash_table.DataTable(
                    id="edge-selection",
                    page_size=20,
                    style_table={
                        "overflowX": "scroll",
                        "overflowY": "scroll",
                        "height": "50vh",
                        "width": "95%"
                    },

                    style_cell_conditional=[
                        {
                            "if": {"column_id": "interaction"},
                            "textAlign": "left"
                        },
                        {
                            "if": {"column_id": "receptorfamily"},
                            "textAlign": "left"
                        },
                        {
                            "if": {"column_id": "pubmedid"},
                            "textAlign": "left"
                        }
                    ],

                    style_header={
                        "fontWeight": "bold",
                        "maxWidth": "200px",
                        "minWidth": "70px"
                    },

                    style_data={
                        "maxWidth": "200px",
                        "minWidth": "70px",
                        "textOverflow": "ellipsis"
                    },
                    sort_action="native",

                    fixed_rows={'headers': True, 'data': 0}
                )
            ])
        ]),  # end interaction list

        html.Div(className="L-R-scores", children=[  # ligand and receptor lists
            html.H2("Ligand and receptors", style={"text-align": "center"}),
            html.Div(children=[
                html.H3(id="selected-node", style={"text-align": "center"}, children=["Select a node in the notwork graph"]),
                html.Label("Search for ligands and receptors:", style={"margin-right": "10px"}),
                dcc.Input(id="filter_l_r", type="search", value="", placeholder="Search")
            ]),
           
           
            dcc.Tabs([
                dcc.Tab(label="Ligands", children=[
                    dcc.Graph(
                        id="ligand-graph",
                        config=dict(
                            autosizable=True,
                            responsive=True)
                    ),
                    dash_table.DataTable(
                    id="ligand-table",
                    page_size=20,
                    style_table={
                        "overflowX": "scroll",
                        "overflowY": "scroll",
                        "height": "50vh",
                        "width": "95%"
                    },

                    style_cell_conditional=[
                        {
                            "if": {"column_id": "Ligand"},
                            "textAlign": "left"
                        }
                    ],

                    style_header={
                        "fontWeight": "bold",
                        "maxWidth": "200px",
                        "minWidth": "70px"
                    },

                    style_data={
                        "maxWidth": "200px",
                        "minWidth": "70px",
                        "textOverflow": "ellipsis"
                    },
                    sort_action="native",

                    fixed_rows={'headers': True, 'data': 0})
                ]),
                dcc.Tab(label="Receptors", children=[
                    dcc.Graph(
                        id="receptor-graph",
                        config=dict(
                            autosizable=True,
                            responsive=True)
                    ),
                    dash_table.DataTable(
                    id="receptor-table",
                    page_size=20,
                    style_table={
                        "overflowX": "scroll",
                        "overflowY": "scroll",
                        "height": "50vh",
                        "width": "95%"
                    },

                    style_cell_conditional=[
                        {
                            "if": {"column_id": "Receptor"},
                            "textAlign": "left"
                        }
                    ],

                    style_header={
                        "fontWeight": "bold",
                        "maxWidth": "200px",
                        "minWidth": "70px"
                    },

                    style_data={
                        "maxWidth": "200px",
                        "minWidth": "70px",
                        "textOverflow": "ellipsis"
                    },
                    sort_action="native",

                    fixed_rows={'headers': True, 'data': 0})
                ])
            ])
        ])  # end ligand receptor list
    ])  # end wrapper

    # Instantiate the graph and produce the bounderies for filters
    @app.callback([
        Output("cyto-graph", "elements"),
        Output("network-filter", "min"),
        Output("network-filter", "max"),
        Output("network-filter", "value")
        ],
        [
        Input("network-interaction", "value"),
        Input("weight-select", "value")])
    def make_graph(interaction, score):
        G_flat = G_to_flat(G, score)
            

        if interaction == "all":  # if no interaction is selected, use full graph
            G_cyto = nx.cytoscape_data(G_flat)
            weights = [d["weight"] for u, v, d in G_flat.edges(data=True)]

            # prepare data for network graph
            nodes = G_cyto["elements"]["nodes"]
            edges = G_cyto["elements"]["edges"]
            elements = nodes+edges

            return elements, min(weights), max(weights), np.mean(weights)

        else:  # an interaction is selected, select only that interaction
            G_split = cn.graph.split_graph(G)
            G_split_flat = G_to_flat(G_split[interaction], score)
            G_cyto = nx.cytoscape_data(G_split_flat)
            weights = [d["weight"]
                       for u, v, d in G_split_flat.edges(data=True)]

            # prepare data for network graph
            nodes = G_cyto["elements"]["nodes"]
            edges = G_cyto["elements"]["edges"]
            elements = nodes+edges

            return elements, min(weights), max(weights), np.mean(weights)

    # Change layout of network graph

    @app.callback(Output("cyto-graph", "layout"),
                  [Input("network-layout", "value")])
    def update_network_layout(layout):
        return {
            "name": layout,
            "automate": True,
            "fit": True
        }

    # Choose gene to color nodes by

    @app.callback(
        [Output("node-colors", "children"),
         Output("min-max", "children")],
        [
            Input("gene_dropdown", "value")
        ]
    )
    def calculate_colors(gene):
        if gene is None:
            return [None, ""]
        # get all gene expression values for selected gene
        gene_data = {celltype["data"]["id"]: celltype["data"]
                     ["genes"][gene] for celltype in nodes}

        min_value = min(gene_data.values())
        max_value = max(gene_data.values())

        # package min max expression information to a list that will be returned
        expression = html.Ul(children=[
            html.Li(f"minimum gene expression: {min_value}"),
            html.Li(f"maximum gene expression: {max_value}")
        ]
        )

        cmap = matplotlib.cm.get_cmap("coolwarm")

        color_dict = dict()
        for k, v in gene_data.items():
            color_dict[k] = {"rgb": cmap(v, bytes=True)[0:3], "expression": v}

        color = pd.Series(color_dict)

        return color.to_json(), expression

    # Select visible edges of network graph depending on filter value
    # node color depending on selected gene
    # width of edges

    @app.callback(
        Output("cyto-graph", "stylesheet"),
        [
            Input("network-filter", "value"),
            Input("network-filter", "min"),
            Input("network-filter", "max"),
            Input("node-size", "value"),
            Input("node-colors", "children")
        ]
    )
    def style_network_graph(th, min_weight, max_weight, size, colors):

        # create a filter for edges
        filter_style = [{
            "selector": f"edge[weight < {th}]",
            "style": {
                "display": "none"
            }
        },
            {
            "selector": "node",
            "style": {
                'height': f'mapData(centrality, {min(cent)}, {max(cent)}, {size[0]}, {size[1]})',
                'width': f'mapData(centrality, {min(cent)}, {max(cent)}, {size[0]}, {size[1]})'
            }
        }]

        # create a color style for nodes based on gene expression
        if isinstance(colors, str):
            colors = pd.read_json(colors, typ="series", convert_dates=False)
            color_style = [{'selector': f'node[id = "{str(index)}"]', 'style': {
                'background-color': f'rgb{tuple(colors[index]["rgb"])}'}} for index in colors.index]
            filter_style += color_style
        else:
            color_style = {
                "selector": "node",
                "style": {'background-color': 'BFD7B5'}
            }

        # Map edges width to a set min and max value (scale for visibility)
        edge_style = [{
            "selector": "edge",
            "style": {
                "width": f"mapData(weight, {min_weight}, {max_weight}, 1, 10)"
            }
        }]

        return default_stylesheet + filter_style + edge_style

    # download an image of current network graph view
    @app.callback(
        Output("cyto-graph", "generateImage"), 
        Input("download-network-graph", "n_clicks"))
    def download_networkgraph_image(get_request):
        
        if get_request == None:
            return dict()

        return {
            "type": "svg",
            "action": "download"
        }


    # Produce a table of all edge data from tapped edge
    @app.callback([
        Output("edge-info", "children"),
        Output("edge-selection", "columns"),
        Output("edge-selection", "data")
    ],
        [Input("cyto-graph", "tapEdgeData"),
        Input("interaction-selection", "children")])
    def update_data(edge, selection):
        import pandas as pd
        import json

        # check if an edge has really been clicked, return default otherwise
        if edge is None:
            return ["", None, None]

        info = f"Interactions from {edge['source']} to {edge['target']}."

        # map visible names for columns with columns in edge[interaction]
        columns = [
            {
                "name": "Interaction",
                "id": "interaction"
            },
            {
                "name": "Receptor Family",
                "id": "receptorfamily"
            },
            {
                "name": "Score",
                "id": "score"
            },
            {
                "name": "Log10(score)",
                "id": "log_score"
            },
            {
                "name": "Specificity",
                "id": "specificity"
            },
            {
                "name": "Importance",
                "id": "importance"
            },
            {
                "name": "Ligand z-score",
                "id": "ligand_zscore"
            },
            {
                "name": "Ligand p-value",
                "id": "ligand_pval"
            },
            {
                "name": "Receptor z-score",
                "id": "receptor_zscore"
            },
            {
                "name": "Receptor p-value",
                "id": "receptor_pval"
            },
            {
                "name": "PubMed ID",
                "id": "pubmedid"
            }
        ]

        interactions = pd.DataFrame(edge["interactions"])[
            ["interaction", "receptorfamily", "score", "log_score", "specificity", "importance", "ligand_zscore",
             "ligand_pval", "receptor_zscore", "receptor_pval", "pubmedid"]]

        # Sort values based on score
        interactions.sort_values(by="score", ascending=False, inplace=True)

        # round values for scores to two decimals
        interactions[[
            "score",
            "log_score",
            "specificity",
            "importance",
            "ligand_zscore",
            "receptor_zscore"]] = interactions[[
                "score",
                "log_score",
                "specificity",
                "importance",
                "ligand_zscore",
                "receptor_zscore"]].round(decimals=2)

        interactions[["ligand_pval", "receptor_pval"]] = interactions[[
            "ligand_pval", "receptor_pval"]].round(decimals=4)

        # if selection from interaction graph, filter dataframe
        if selection != "":
            selection = json.loads(selection)
            interactions = interactions.loc[interactions["interaction"].isin(selection)]

        records = interactions.to_dict("records")

        return [info, columns, records]


    @app.callback([
        Output("interaction-scatter", "figure")
    ],
        [Input("cyto-graph", "tapEdgeData")])
    def interaction_scatter_plot(edge):
        import plotly.express as px

        fig = go.Figure()
        if not isinstance(edge, dict):
            return [fig, ]

        interactions = pd.DataFrame(edge["interactions"])[
                    ["interaction", "receptorfamily", "score", "log_score", "ligand_zscore",
                    "ligand_pval", "receptor_zscore", "receptor_pval", "specificity", "importance", "pubmedid"]]
        
        # add 10% to the min and max value to not clip the datapoint
        range_x = (-max(interactions["log_score"])*0.1, max(interactions["log_score"])*1.1)
        range_y = (-max(interactions["specificity"])*0.1, max(interactions["specificity"])*1.1)
        #interactions["specificity"] = np.log10( interactions["specificity"])

        fig = px.scatter(interactions, 
                x="log_score",
                range_x=range_x,
                y="specificity",
                range_y=range_y,
                color="importance",
                hover_name="interaction",
                hover_data=["ligand_pval", "receptor_pval", "score","specificity", "receptorfamily"],
                color_continuous_scale=px.colors.sequential.Viridis_r,
                labels={
                    "ligand_zscore": "Ligand Z-score",
                    "receptor_zscore": "Receptor Z-score",
                    "log_score": "log(Interaction score)",
                    "score": "Interaction score",
                    "specificity": "Specificity",
                    "importance": "Importance",
                    "receptorfamily": "Receptor family",
                    "pubmedid": "PubMed ID",
                    "ligand_pval": "Ligand p-value",
                    "receptor_pval": "Receptor p-value"
                    }
                )
        return [fig,]

    @app.callback(
        Output("interaction-selection", "children"),
        [
        Input("interaction-scatter", "selectedData")
        ]
        )
    def interaction_select(selected_data):
        import json
        if isinstance(selected_data, dict):
            interactions = [point["hovertext"] for point in selected_data["points"]]
        else:
            return ""
        return json.dumps(interactions)



    # Produce ligand and receptor graphs based on tapped node

    @app.callback([
        Output("ligand-graph", "figure"),
        Output("receptor-graph", "figure"),
        Output("selected-node", "children")
    ],
        [
        Input("cyto-graph", "tapNodeData"),
        Input("filter_l_r", "value")
    ]
    )
    def plot_l_r_expression(node, filter_text):

        # set output variables to empty figures
        ligand_fig = go.Figure()
        receptor_fig = go.Figure()
        node_id = "Select a node in the network graph"

        if isinstance(node, dict):
            import plotly.express as px

            node_id = node["id"]
            
            ligands_score = pd.DataFrame.from_dict(node["ligands_score"], orient="index", columns=["Score"])
            ligands_zscore = np.log2(pd.DataFrame.from_dict(node["ligands_zscore"], orient="index", columns=["Z-score"]))
            ligands_corr_pval = pd.DataFrame.from_dict(node["ligands_corr_pval"], orient="index", columns=["p-value"])
            ligands_merge = ligands_score.merge(ligands_zscore, how="left", left_index=True, right_index=True)
            ligands_merge = ligands_merge.merge(ligands_corr_pval, how="left", left_index=True, right_index=True)
            ligands_merge["log(score + 1)"] = np.log10(ligands_merge["Score"]+1)
            ligands_merge["Significant"] = [True if p_val < 0.05 else False for p_val in ligands_merge["p-value"]]
            ligands_merge["-log(p-value)"] = -np.log10(ligands_merge["p-value"])
            
            if filter_text != "":
                ligands_merge = ligands_merge.filter(like=filter_text, axis=0)

            ligand_fig = px.scatter(
                ligands_merge, 
                x="log(score + 1)", 
                y="-log(p-value)",
                color="Significant",
                hover_name=ligands_merge.index,
                hover_data=["Score", "Z-score", "p-value"])

            receptors_score = pd.DataFrame.from_dict(node["receptors_score"], orient="index", columns=["Score"])
            receptors_zscore = np.log2(pd.DataFrame.from_dict(node["receptors_zscore"], orient="index", columns=["Z-score"]))
            receptors_corr_pval = pd.DataFrame.from_dict(node["receptors_corr_pval"], orient="index", columns=["p-value"])
            receptors_merge = receptors_score.merge(receptors_zscore, how="left", left_index=True, right_index=True)
            receptors_merge = receptors_merge.merge(receptors_corr_pval, how="left", left_index=True, right_index=True)
            receptors_merge["log(score + 1)"] = np.log10(receptors_merge["Score"]+1)
            receptors_merge["Significant"] = [True if p_val < 0.05 else False for p_val in receptors_merge["p-value"]]
            receptors_merge["-log(p-value)"] = -np.log10(receptors_merge["p-value"])
            
            if filter_text != "":
                receptors_merge = receptors_merge.filter(like=filter_text, axis=0)

            receptor_fig = px.scatter(
                receptors_merge, 
                x="log(score + 1)", 
                y="-log(p-value)",
                color="Significant",
                hover_name=receptors_merge.index,
                hover_data=["Score", "Z-score", "p-value"])
       
        return [ligand_fig, receptor_fig, node_id]

    # Builds a sankey graph based on the tapped node (store in global G_s)
    G_s = nx.MultiDiGraph() #variable holding sankey graph
    @app.callback([
        Output("sankey-filter", "min"),
        Output("sankey-filter", "max"),
        Output("sankey-filter", "value")
        ],
        [
        Input("cyto-graph", "tapNodeData"),
        Input("sankey-toggle", "value")])
    def build_sankey_graph(node, score):
        import numpy as np
        # If no node has been selected, dont try to build graph
        if node is None:
            return (0 ,0 ,0)

        node = node["id"]
        # Find all interactions where node is target or source node
        nonlocal G_s
        G_s = nx.MultiDiGraph() # reset content
        weight = list() # list to store all weights (used to set min and max for the filter)
        for n, nbrs in G.adj.items(): # graph has been modified by network graph before
            for nbr, edict in nbrs.items():
                if n == node:
                    for e, d in edict.items():
                        G_s.add_edge(n, " Post " + nbr, **d)
                        weight.append(d[score])
                if nbr == node:
                    for e, d in edict.items():
                        G_s.add_edge("Pre " + n, nbr, **d)
                        weight.append(d[score])

        if len(weight) == 0:
            weight = [0,1]
        if score == "specificity": 
            # set default start value to specificity value for ligand and receptor 
            # p-value of (0.05 and 0.05)/2 = 1.3
            return (min(weight), max(weight), 1.3)
        return (min(weight), max(weight), np.mean(weight))
    
    @app.callback(
        Output("sankey-graph", "figure"),
        [Input("sankey-filter", "value"),
        Input("sankey-toggle", "value"),
        Input("cyto-graph", "tapNodeData")])
      
    def filter_sankey_graph(th, score, node):

        if node:
            node = node["id"]

        _G_s = nx.MultiDiGraph()
        for u, v, n, d in G_s.edges(data=True, keys=True):
            if d[score] > th:
                _G_s.add_edge(u, v, n, **d)
        _G_s.add_nodes_from(G_s.nodes(data=True))
        
        edges = nx.to_pandas_edgelist(_G_s)
        if len(edges) < 1:
            fig = dict()
            return fig
        # add same color scheme as network graph
        for node_s in _G_s.nodes():
            if " Post" in node_s:
                original_node = str(node_s).split(sep=" Post")[1]
            elif "Pre " in node_s:
                original_node = str(node_s).split(sep="Pre ")[1]
            else:
                original_node = str(node_s)

            new_color = color_map_nodes[original_node.strip()]
            G_s.nodes[node_s]["color"] = new_color

        nodes = G_s.nodes()

        node_map = {cluster: id for id, cluster in enumerate(list(nodes))}

        sankey = go.Sankey(
            node=dict(
                pad=15,
                thickness=20,
                line=dict(
                    color="black",
                    width=0.5
                ),
                label=list(nodes),
                color=[f'rgb{tuple(d["color"][0:3])}' for n,
                       d in G_s.nodes(data=True)]

            ),
            link=dict(
                source=list(edges["source"].map(node_map)),
                target=list(edges["target"].map(node_map)),
                value=list(edges[score]),
                label=edges["interaction"]
            )
        )

        data = [sankey]

        layout = go.Layout(
            autosize=True,
            title=f"Interactions: {node}",
            font=dict(
                size=font_size
            )
        )

        fig = go.Figure(data=data, layout=layout)

        return fig

    @app.callback([Output("ligand-table", "columns"), Output("ligand-table", "data")], [Input("ligand-graph", "figure"), Input("ligand-graph", "selectedData")])
    def select_ligands(figure, selected):
        import json
        ligands = []
        score= []
        zscore = []
        pval = []
        
        

        for group in figure["data"]:
            for ligand in group["hovertext"]:
                ligands.append(ligand)
            for data in group["customdata"]:
                score.append(data[0])
                zscore.append(data[1])
                pval.append(data[2])

        
        df = pd.DataFrame({"Ligand": ligands, "Score": score, "Z-score": zscore, "P-value": pval})
        df.index = df["Ligand"]
        df.sort_values(by="Score", ascending=False, inplace=True)

        if isinstance(selected, dict):
            filt=[]
            for point in selected["points"]:
                filt.append(point["hovertext"])
            df = df.loc[filt]

        
        columns = [
            {"name": "Ligand", "id": "Ligand"},
            {"name": "Score", "id": "Score"},
            {"name": "Z-score", "id": "Z-score"},
            {"name": "P-value", "id": "P-value"}]

        data = df.to_dict("records")

        return columns, data

    @app.callback([Output("receptor-table", "columns"), Output("receptor-table", "data")], [Input("receptor-graph", "figure"), Input("receptor-graph", "selectedData")])
    def select_ligands(figure, selected):
        import json
        receptors = []
        score= []
        zscore = []
        pval = []
        
        

        for group in figure["data"]:
            for receptor in group["hovertext"]:
                receptors.append(receptor)
            for data in group["customdata"]:
                score.append(data[0])
                zscore.append(data[1])
                pval.append(data[2])

        
        df = pd.DataFrame({"Receptor": receptors, "Score": score, "Z-score": zscore, "P-value": pval})
        df.index = df["Receptor"]
        df.sort_values(by="Score", ascending=False, inplace=True)


        if isinstance(selected, dict):
            filt=[]
            for point in selected["points"]:
                filt.append(point["hovertext"])
            df = df.loc[filt]

        
        columns = [
            {"name": "Receptor", "id": "Receptor"},
            {"name": "Score", "id": "Score"},
            {"name": "Z-score", "id": "Z-score"},
            {"name": "P-value", "id": "P-value"}]

        data = df.to_dict("records")

        return columns, data

    # Run server
    app.run_server(**kwargs)


if __name__ == "__main__":

    import scConnect as cn
    print("Loading graph file....")
    G = cn.graph.load("assets/graph.yaml")
    print("graph file loaded.")
    G = cn.graph.score(G)
    #G = "test"
    graph(G)



import scConnect as cn
version = cn.database.version
organism = cn.database.organism

def build_graph(edge_list, node_list=False):
    """Construct a multi directional graph from an edgelist and nodelist and save as graph.
    Also return the graph object as a nx.MultiDiGraph.

    If no nodelist is passed, nodes will be implied from edgelist. 
    Create a nodelist using :py:func:`scConnect.connect.nodes`
    """
    import networkx as nx
    G = nx.MultiDiGraph()
    G.add_edges_from([(u, v, d) for u, v, d in edge_list])

    if isinstance(node_list, list):
        try:
            G.add_nodes_from(node_list)
        except:
            pass
    print(
        f"Graph has {G.number_of_edges()} interactions between {G.number_of_nodes()} clusters")

    nodes = list(G.nodes())
    for node in nodes:
        print(f"{node} has {G.degree(node)} interactions")

    # Calculate all edge and node metrics
    G = score(G)

    return G


# Simple function to abstact away some networkX API
def save(G, file_name):
    """Save the graph to .yaml format

    .. note::
        Yaml was the only format that supported utf-8 attributes and dictionary attributes
    """
    import networkx as nx
    nx.write_yaml(G, file_name, encoding="utf-8")


def load(filename):
    """load the graph from yaml format"""
    import networkx as nx
    G = nx.read_yaml(filename)
    return G
# Here follows functions to calculate edge atributes to help fins specifi interactions


def promiscuity(G):
    """
    Calculates edge attribute measure based on 
    the number of clusters a ligand is used to connect with.

    This measure is ligand specific.
    """
    ligand_count = dict()
    for n, nbrs in G.adj.items():
        cluster_list = dict()
        for nbr, edict in nbrs.items():
            # We do only want to check if a ligand is connected to the neighbor once.
            ligand_check = dict()
            for e, d in edict.items():
                if d["ligand"] not in ligand_check.keys():
                    if d["ligand"] in cluster_list.keys():
                        cluster_list[d["ligand"]] += 1
                    else:
                        cluster_list[d["ligand"]] = 1
                    # Lets stop counting this ligand
                    ligand_check[d["ligand"]] = "counted for this neighbor"
        ligand_count[n] = cluster_list
    # Adds normalized promiscuity value for each edge depending on ligand used.
    tot_n = len(G.nodes())
    for n, nbrs in G.adj.items():
        for nbr, edict in nbrs.items():
            for e, d in edict.items():
                G.edges[n, nbr, e]["promiscuity"] = ligand_count[n][d["ligand"]]/tot_n

    return G


def loyalty(G):
    """
    calculates edge attribute measure based on ratio of interaction
    to the neighbor over total interactions to all neighbors

    this measure is neighbour specific.
    """
    score_dict = dict()
    for n, nbrs in G.adj.items():
        scores = dict()
        for nbr, edict in nbrs.items():
            scores[nbr] = len(edict)
        # normalize on total interactions
        total = sum(scores.values())
        for nbr in scores.keys():
            scores[nbr] /= total
        score_dict[n] = scores

    # Adds the loyalty score to each edge depending on neigbour
    for n, nbrs in G.adj.items():
        for nbr, edict in nbrs.items():
            for e in edict.keys():
                G.edges[n, nbr, e]["loyalty"] = score_dict[n][nbr]
    return G


def weighted_score(G):
    """
    Calculate a weighted score based on original score,
    loyalty and promiscuity for each interaction.

    High loyalty promotes interactions between a set of neighbours
    High promiscuity punish interactions of specific ligands.
    """
    from math import log1p
    from numpy import log10
    for n, nbrs in G.adj.items():
        for nbr, edict in nbrs.items():
            for e, d in edict.items():
                # punnish interaction score low loyalty and for high promiscuity
                weighted_score = d["score"] * d["loyalty"] / d["promiscuity"]
                # Add w_score to the edge attribute
                G.edges[n, nbr, e]["weighted_score"] = weighted_score
                G.edges[n, nbr, e]["log_weighted_score"] = log10(weighted_score)
    return G


def centrality(G):
    """
    Calculate information centrality for all nodes in the graph

    returns a graph with added node attributes"""
    import networkx as nx

    cent = nx.degree_centrality(G)
    for node in G.nodes():
        G.nodes[node]["centrality"] = cent[node]

    return G


# Summary function for all graph measures (used to quickly run all)


def score(G):
    """
    Calculate all edge and node attribute measures for interactions in the graph.

    This needs to be run before flatten graph and is automatically
    run when constructing a graph with :py:func:`scConnect.graph.build_graph`
    """
    # Edge attributes
    G = promiscuity(G)
    G = loyalty(G)
    G = weighted_score(G)

    # Node attributes
    G = centrality(G)
    # To be added if needed
    return G

# The flatten graph aproach has been depreceated.
# Use conversion to adjacency matrix on filtered graphs insted

# filter the graph based on interactions containing ligands and receptors under a threshold p-value
def significant_interactions(G, a=0.05):
    """Return a filtered graph containing only sigificanly differentiated ligands and receptor interaction.
    
    a: alpha, selects scores with a pvalue under alpha.
    Return a MultiDiGraph
    """

    import networkx as nx
    G_filtered = nx.MultiDiGraph()
    for u, v, n, d in G.edges(data=True, keys=True):
        # Find interactions with significant p-values (pval < a)
        # Must be significantly upregulated (zscore > 0)
        if d["ligand_pval"] < a and d["receptor_pval"] < a and d["ligand_zscore"] > 0 and d["receptor_zscore"] > 0:
            G_filtered.add_edge(u, v , key=n ,**d)
            
    G_filtered.add_nodes_from(G.nodes(data=True))
    
    return G_filtered
    


def flatten_graph(G, weight="score", log=True):
    """
    Collapse graph and calculate edge weight between clusters as:
    weight = sum(weighted_score)

    log: log the sum of scores

    Returns: a nx.DiGraph

    .. note::
        Use flatten graph to export to a simpler format, for instance
        if working with gephi or other softwares.
        Note that edge atributes for specific interactions are
        stored as a list of dictionaries, but might be 
        unaccessible for most programs.

    """
    import networkx as nx
    from numpy import log1p, log10
    G_flat = nx.DiGraph()
    for n, nbrs in G.adj.items():
        for nbr, edict in nbrs.items():
            interaction_scores = list()  # collect all weighted interaction scores in this list
            for e, d in edict.items():
                interaction_scores.append(edict[e][weight])

            weight_sum = sum(interaction_scores)
            if log:
                weight_sum = log10(weight_sum +1)                

            # Add edge and pass in edict for possibility to access later
            G_flat.add_edge(n, nbr, weight=weight_sum,
                            interactions=list(edict.values()))

    # Forward node attributes
    G_flat.add_nodes_from(G.nodes(data=True))
    return G_flat


# Functions to access underlying interactions in a flattened graph
def get_interactions(G, n, nbr):
    """
    Find interactions between nodes in a flattened graph
    and returns a list sorted on score
    """
    edge = G[n][nbr]
    import pandas as pd
    interactions = pd.DataFrame(edge["interactions"])
    interactions.sort_values("score", ascending=False, inplace=True)

    return interactions


def get_emmitors(G, n):
    """
    Find all clusters emitting to selected cluster and
    returns a dictionary of emittor weight pairs
    """
    import networkx as nx
    emittors = {emittor: data["weight"] for (
        emittor, target, data) in G.edges(data=True) if target == n}
    return emittors


def get_targets(G, n):
    """
    Find all clusters targeted by selected cluster and
    returns a dictionary of target weight pairs
    """
    import networkx as nx
    targets = {target: data["weight"] for (
        emittor, target, data) in G.edges(data=True) if emittor == n}
    return targets


def flatten_graph_to_adjacency_matrix(G, weight="Weighted_score", log=False):
    """flattens graph over provided weight and returns a sorted adjacency matrix
    if log = True, performs a log10 transformation of the results
    """
    import networkx as nx
    import numpy as np
    import pandas as pd

    if len(G.edges()) == 0:
        matrix = np.zeros(shape=(len(G.nodes()), len(G.nodes())))
        matrix = pd.DataFrame(matrix, columns=G.nodes(), index=G.nodes())
    else:
        matrix = nx.to_pandas_adjacency(G, weight=weight)
        ind = matrix.index.values
        matrix = matrix.sort_index(ascending=True)
        matrix = matrix[matrix.index.values]
        matrix = pd.DataFrame(matrix, columns=ind, index=ind)

    if log:
        matrix = np.log10(matrix)
    return matrix


def split_graph(G):
    """splits graph(s) on interactions and return a dictionary of graphs with interaction as keys."""
    import networkx as nx

    # Find all interactions to split the graph on
    split_by = "interaction"
    split_list = list()
    for u, v, d in G.edges(data=True):
        split_list.append(d[split_by])
    split_set = set(split_list)
    G_splits = dict()
    for split in split_set:
        G_split = nx.from_edgelist(
            [(u, v, d) for u, v, d in G.edges(data=True) if d[split_by] == split],
            create_using=nx.MultiDiGraph())

        G_split.add_nodes_from(G.nodes(data=True))
        G_splits[split] = G_split

    return G_splits


def plot_adjacency_matrix(G, save=False, filename=None, weight="weighted_score", log=False, width=200, height=200, fontsize=10, fontscale=1.5):
    """Flattens and plots an adjecency matric for a given graph.
    If a split graph is provided, plots adjacency matrix for each interaction.

    Weight can be "weighted_score" or "score".
    Returns: Holoviews heatmap
    """
    import numpy as np
    import holoviews as hv
    from holoviews import opts
    import networkx as nx
    hv.extension('bokeh')

    # Create adjecency matrix and sort index and columns by name
    matrix = nx.to_pandas_adjacency(G, weight=weight)

    if log:
        matrix = np.log10(matrix)

    matrix = matrix.sort_index(ascending=True)
    matrix = matrix[matrix.index.values]
    matrix = matrix.stack()
    matrix = matrix.reset_index(name="values")
    matrix.columns = ["source", "target", "values"]

    # Plot the matrix as a heatmap
    heatmap = hv.HeatMap(matrix, kdims=["target", "source"])

    # Rotate X lables and show the color bar
    heatmap.opts(
        opts.HeatMap(tools=["hover"], xrotation=90, colorbar=True, width=width, height=height, fontsize=fontsize, fontscale=fontscale)
    )

    if save is True:
        hv.save(heatmap, filename)
    return heatmap


def corr_adjacency_matrix(G1, G2, weight="score"):
    """finds the pearson correlation between two adjecency matrix.

    Returns: Pearson correlation coeficient"""
    import networkx as nx
    import numpy as np

    # Create adjacency matric from Graph 1 and sort index and columns
    # by name and flatten the matrix to an array
    m1 = nx.to_pandas_adjacency(G1, weight=weight)
    m1 = m1.sort_index(ascending=True)
    m1 = m1[m1.index.values]
    m1 = m1.values.flatten()

    # Create adjacency matric from Graph 2 and sort index and columns
    # by name and flatten the matrix to an array
    m2 = nx.to_pandas_adjacency(G2, weight=weight)
    m2 = m2.sort_index(ascending=True)
    m2 = m2[m2.index.values]
    m2 = m2.values.flatten()

    # Find the pearson correlation coefficiant between the two arrays
    corr = np.corrcoef(m1, m2)[0][1]
    return corr


def plot_chord(G, notebook=False):
    """
    render an interactive chord plot of the graph.

    If notebook is False, starts an bokeh app in a browser window, 
    if True, renders the plot directly in the cell.
    """

    import scConnect as cn
    import holoviews as hv
    from holoviews import opts, dim
    import networkx as nx
    import pandas as pd
    import numpy as np

    # instantiate the bokeh renderer
    renderer = hv.renderer('bokeh')
    hv.extension("bokeh")
    hv.output(size=250)

    # set visuals
    opts.defaults(
        opts.Chord(
            node_cmap='Category20',
            edge_cmap='Category20',
            edge_color=dim("source"),
            labels='cluster',
            node_color=dim('cluster'),
            inspection_policy="edges",
            toolbar="above",
        )
    )

    # Creat the Dataset object to be passed to the Chord object (NOTE: aggregates the data
    # leaving only one edge per cluster pair)
    edges = nx.to_pandas_edgelist(G)
    links = hv.Dataset(edges, ["source", "target"], ["weighted_score", "loyalty"]).sort(
        by="source").aggregate(function=np.sum)
    nodes = hv.Dataset(list(G.nodes), 'cluster').sort(by="cluster")

    # Calculate values for threshold representing given percentiles
    percentiles = [0, 20, 40, 60, 80, 90, 95, 99]
    th_values = np.percentile(links.data["weighted_score"], percentiles)

    th = hv.Dimension(("th", "weighted scores threshold"),
                      default=th_values[0])

    # Filter data on threshold, and return a chord element
    def chord_graph(th):
        links_ = links.select(weighted_score=(th, None))
        chord = hv.Chord((links_, nodes))

        return chord

    # instantiate the dynamic map
    chord = hv.DynamicMap(chord_graph, kdims=[th]).redim.values(th=th_values)

    # Run the server if not in notebook
    if notebook == False:
        server = renderer.app(chord, show=True, new_window=True)

    if notebook == True:
        return chord


def plot_sankey(G, notebook=False):
    """
    render an interactive sankey plot of the graph.

    If notebook is False, starts an bokeh app in a browser window, 
    if True, renders the plot directly in the cell.
    """

    import scConnect as cn
    import holoviews as hv
    from holoviews import opts, dim
    import networkx as nx
    import pandas as pd
    import numpy as np

    # instantiate the bokeh renderer
    renderer = hv.renderer('bokeh')
    hv.extension("bokeh")
    hv.output(size=100)

    # set visuals
    opts.defaults(
        opts.Sankey(
            cmap='Category20',
            edge_cmap='Category20',
            edge_color=dim("receptorfamily"),
            labels='cluster',
            node_color=dim('cluster'),
            inspection_policy="edges",
            selection_policy="edges",
            colorbar=True,
            toolbar="above"
        ),
        opts.Bars(
            invert_axes=False,
            xrotation=70,
            toolbar="above"
        )
    )

    # Create values to be used for filtering of the graph
    edges = nx.to_pandas_edgelist(G)
    percentiles = [0, 20, 40, 60, 80, 90, 95, 99]
    th_values = np.percentile(edges["weighted_score"], percentiles)
    nodes_list = list(G.nodes())

    node = hv.Dimension(("node", "Focus cluster"), default=nodes_list[0])
    th = hv.Dimension(('th', 'Weighted score threshold'), default=th_values[0])

    # Filter data on node and threshold, and return a sankey element
    def sankey_graph(node, th):
        # Find all interactions where node is target or source node
        G_s = nx.MultiDiGraph()
        for n, nbrs in G.adj.items():
            for nbr, edict in nbrs.items():
                if n == node:
                    for e, d in edict.items():
                        # append dash after the target node
                        G_s.add_edge(n, nbr + "_", **d)
                if nbr == node:
                    for e, d in edict.items():
                        # append dash before the source node
                        G_s.add_edge("_" + n, nbr, **d)
        # create the dataset used to build the sankey graph.
        # Sort values on weight to get ordered representation on plot.
        edges = nx.to_pandas_edgelist(G_s)
        links = hv.Dataset(edges, ["source", "target"],
                           ["weighted_score", "interaction", "receptorfamily"]).sort("weighted_score")
        nodes = hv.Dataset(list(G_s.nodes), 'cluster')
        sankey = hv.Sankey((links, nodes)).select(weighted_score=(th, None))

        # calculate bars
        ligands = hv.Dataset(edges, ["ligand", "source"], ["score"]).select(
            source=node).aggregate(function=np.mean).sort("score", reverse=True)
        receptors = hv.Dataset(edges, ["receptor", "target"], ["score"]).select(
            target=node).aggregate(function=np.mean).sort("score", reverse=True)
        bars = (hv.Bars(ligands, "ligand") +
                hv.Bars(receptors, "receptor")).cols(2)

        # calculate table
        ligands = hv.Dataset((G.node[node]["ligands"]), "ligand", "score").sort(
            "score", reverse=True)
        receptors = hv.Dataset((G.node[node]["receptors"]), "receptor", "score").sort(
            "score", reverse=True)
        table = hv.Layout(hv.Table(ligands) + hv.Table(receptors)).cols(2)

        return (bars + table + sankey).cols(2)

    sankey = hv.DynamicMap(sankey_graph, kdims=[node, th]).redim.values(
        node=nodes_list, th=th_values)

    layout = sankey
    # Run the server if not in notebook
    if notebook == False:
        server = renderer.app(layout, show=True, new_window=True)

    if notebook == True:
        return layout




def compare_interactions_df(G, node_a=str, node_b=str, method="ratio"):
    
    import networkx as nx
    import pandas as pd
    import numpy as np
    
    Gs = split_graph(G)
    interaction_dict_i = dict()
    for node in G.nodes():
        ratio_dict = dict()
        for interaction in Gs:
            
            if method == "ratio":
                df = nx.to_pandas_adjacency(Gs[interaction], weight="log_score")
                a = df[node_a][node]
                b = df[node_b][node]
                ratio = a - b   # calculate the ratio of interaction scores (using log values log(a)-log(b) = log(a/b))
                                # This bypass divition with 0 issues as in a/b, b might be 0

            if method == "difference":
                df = nx.to_pandas_adjacency(Gs[interaction], weight="score")
                a = df[node_a][node]
                b = df[node_b][node]
                ratio = a - b  # Calulate the difference in interaction score

            if method == "specificity":
                df = nx.to_pandas_adjacency(Gs[interaction], weight="specificity")
                a = df[node_a][node]
                b = df[node_b][node]
                ratio = a - b  # Calulate the difference in specificity

            ratio_dict[interaction] = ratio
        interaction_dict_i[node] = ratio_dict
    df_i = pd.DataFrame(interaction_dict_i)


    interaction_dict_o = dict()
    for node in G.nodes():
        ratio_dict = dict()
        for interaction in Gs:
            if method == "ratio":
                df = nx.to_pandas_adjacency(Gs[interaction], weight="log_score")
                a = df[node][node_a]
                b = df[node][node_b]
                ratio = a - b   # calculate the ratio of interaction scores (using log values log(a)-log(b) = log(a/b))
                                # This bypass divition with 0 issues as in a/b, b might be 0
                
            if method == "difference":
                df = nx.to_pandas_adjacency(Gs[interaction], weight="score")
                a = df[node][node_a]
                b = df[node][node_b]
                ratio = a - b  # Calulate the difference in interaction score
                
            ratio_dict[interaction] = ratio
        interaction_dict_o[node] = ratio_dict
    df_o = pd.DataFrame(interaction_dict_o)
    return df_i, df_o




def compare_interactions_plot(G, node_a=str, node_b=str, th=float, figsize_i=(5, 15), figsize_o=(5,15), colormap="nipy_spectral", method="ratio", save=None):
    """Compares interactions scores between node a <--> x and node b <--> x. returns the logaritmized ratio.
    
    if method = "ratio": log(a/b) = log(a)-log(b), "th": times difference ie. 4x
                "difference": values = a-b, "th": difference ie. 200
     

    returns two plots, first incomming interactions to a an b, then outgoing interaction from a and b. 
    also a lut and color pallet for the receptor families
    """
    import seaborn as sns
    import numpy as np
    import scConnect as cn
    import matplotlib.pyplot as plt
    
    df_i, df_o = compare_interactions_df(G, node_a, node_b, method)

    if method == "ratio": # convert a ratio th to the corresponding log values
        th = np.log10(th)


    # Filter interactions
    interaction_filter_i = [True if max(abs(values)) > th else False for i, values in df_i.iterrows()]
    df_i = df_i[interaction_filter_i]
    interaction_filter_o = [True if max(abs(values)) > th else False for i, values in df_o.iterrows()]
    df_o = df_o[interaction_filter_o]

    # Find receptors used by both incomming and outgoing interactions
    receptors_i = [interaction.split("--> ")[1] for interaction in df_i.index]
    receptors_o = [interaction.split("--> ")[1] for interaction in df_o.index]
    receptors = set(receptors_i + receptors_o)
    
    # find receptor families for all selected interactions to color rows
    targets = cn.database.get_data("targets")
    targets = targets.set_index("Target name")

    # create a dictionary of receptors and families and a correspinding lut
    families_dict = targets.loc[receptors]["Family name"].to_dict() # selects only families for which we have receptors in the dataset
    unique_families = set(families_dict.values())
    lut = dict(zip(unique_families, sns.color_palette(colormap, len(unique_families))))

    if df_i.shape[0] > 1:
        # Make color mapping for incomming interactions
        row_colors_i = list()
        for receptor in receptors_i:
            family = families_dict[receptor]
            color = lut[family]
            row_colors_i.append(color)

        # plot a clustered heatmap with receptor families as cluster row color for incomming interactions
        print("incomming interactions")

    
    
        sns.clustermap(df_i, center=0, cmap="bwr", figsize=figsize_i, row_colors=row_colors_i)
        if save != None:
            path_i = f"{save}{node_a}_to_{node_b}_incomming.pdf"
            plt.savefig(path_i)

    if df_o.shape[0] > 1:
        # Make color mapping for outgoing interactions
        row_colors_o = list()
        for receptor in receptors_o:
            family = families_dict[receptor]
            color = lut[family]
            row_colors_o.append(color)

        # plot a clustered heatmap with receptor families as cluster row color
        print("Outgoing interactions")
        sns.clustermap(df_o, center=0, cmap="bwr", figsize=figsize_o, row_colors=row_colors_o)
        if save != None:
            path_o = f"{save}{node_a}_to_{node_b}_outgoing.pdf"
            plt.savefig(path_o)

    print(f"receptor families in order: {list(lut.keys())}")
    sns.palplot(lut.values())

    return None


def plot_interaction_heatmap(G, th=0.5, method='ward', metric='euclidean', z_score=None, standard_scale=None, figsize=None, save=None):
    """plots a heatmap of logaritmized interaction scores for each Source-->Target combination. 
    
    
    G: multi-directional graph
    th: filter out all interactions where the maximum interaction score is lower than the threshold (log values)
    method: passed to seaborn.clustermap
    metric: passed to seaborn.clustermap
    z_score: passed to seaborn.clustermap. calculates z-scores for each value x in the matrix as z=(x-mean)/std.
    standard_scale: passed to seaborn.clustermap
    figsize: passed to seaborn.clustermap. this is automatically set based on size of dataset, but can be explicity set using a tuple.
    save: filename with file extension

    return: seaborn.matrix.ClusterGrid object
    """

    import networkx as nx
    import pandas as pd
    import seaborn as sns
    import itertools

    # split graph on interactions
    print("split graph...")
    Gs = split_graph(G)
    
    ams = dict()
    print("flatten interactions...")
    for interaction in Gs.keys():
        g = Gs[interaction]
        am = nx.to_pandas_adjacency(g, weight="log_score") #make adjacency matrix with scores
        am = pd.Series(am.values.flatten()) # flatten the array
        ams[interaction] = am # add flatten adjacency to dict
    
    print("recovering index...")
    #create index
    am = nx.to_pandas_adjacency(Gs[list(Gs.keys())[0]])
    ix = list(itertools.product(am.index.values, repeat=2))
    ix = [f"{pops[0]}-->{pops[1]}" for pops in ix]

    # Create a df with the flatten arrays
    print("build dataframe...")
    df = pd.DataFrame(ams)
    df.index = ix
    filter_interactions = df.aggregate(max, axis=0) > th
    df = df.loc[:,filter_interactions.values]

    print("plotting...")
    if figsize == None:
        figsize = (len(df.index)/3, len(df.columns)/3)
    f = sns.clustermap(df.T, figsize=figsize, method=method, metric=metric, z_score=z_score, standard_scale=standard_scale)
    f.savefig(save)
    
    return f


# Fetch edge list dataframe for plotting functions
def edge_list(
    G, 
    filter_by="specificity", 
    th=1, 
    cellphonedb_interactions=False, 
    interaction_list=None,
    version = version,
    organism = organism,
    all_connections=True):
    """retrieves all interactions from the graphs and returns a pandas dataframe suitable for plotting"""

    import networkx as nx
    import pandas as pd
    import numpy as np
    import plotly.express as px
    import pkg_resources
    
    
    df = nx.to_pandas_edgelist(G)
    
    # change name of scConnects interaction to better compare to cellPhoneDB
    receptor_info = pd.read_csv(pkg_resources.resource_filename("scConnect", (f"data/Gene_annotation/{version}/{organism}/receptors.csv")), index_col=1)
    ligand_info = pd.read_csv(pkg_resources.resource_filename("scConnect", (f"data/Gene_annotation/{version}/{organism}/ligands.csv")), index_col=1)

    df["ligand_gene"] = [eval(ligand_info.loc[ligand]["preprogene"])[0] if isinstance(ligand_info.loc[ligand]["preprogene"], str) else ligand for ligand in df["ligand"]]
    df["receptor_gene"] = [eval(receptor_info.loc[receptor]["gene"])[0] if isinstance(receptor_info.loc[receptor]["gene"], str) else receptor for receptor in df["receptor"]]

    df["cellphonedb_interactions"] = [f"{l}_{r}" for l, r in list(zip(df["ligand_gene"], df["receptor_gene"]))]

    #Create a new connection annotation merging source and target info
    df["connection"] = [str(connection[0]+"|"+connection[1]) for connection in zip(df["source"], df["target"])]

    #Collect all connection for specific interactions
    if all_connections:
        # select all significant interaction (significance > 1)
        interactions = df[df[filter_by] >= th]["interaction"].unique()
        # keep all interactions that were significant between any connection
        df = df[[True if interaction in interactions else False for interaction in df["interaction"]]]

        # only sort if all_connections == True
         # sort based on connection and interactoin names to get order in the plot
        if cellphonedb_interactions:
            df.sort_values(by=[ "cellphonedb_interactions", "connection"], key=np.vectorize(str.lower), ascending=False, inplace=True)
        else:
            df.sort_values(by=["connection", "interaction"], key=np.vectorize(str.lower), ascending=False, inplace=True)
    
    else:
        # if only selected interactions, sort by filter_by key
        df= df[df[filter_by] >= th]
        df.sort_values(by=[filter_by], ascending=False, inplace=True)
   
    return df

# plotting function using the edge list data
def dotplot(
    G=None, 
    df=None, 
    filter_by="specificity", 
    th=1,
    organism=organism,
    version=version,
    all_connections=True,
    cellphonedb_interactions=False, 
    height_scale=40, 
    width_scale=40, 
    cmap=None):

    import plotly.express as px
    import pandas as pd
    import numpy as np

    if cmap == None:
        cmap = px.colors.sequential.Viridis_r
    
    if isinstance(df, pd.DataFrame):
        df = df
        # Filter by feature and threshold
        interactions = df[df[filter_by] >= th]["interaction"].unique()
        # keep all interactions that were significant between any connection
        df = df[[True if interaction in interactions else False for interaction in df["interaction"]]]
    else:
        df = edge_list(G, filter_by, th, organism=organism, version=version, all_connections=all_connections)
        
    # set width and hight of figure
    width = len(df["connection"].unique())*width_scale
    height = len(df["interaction"].unique())*height_scale
    if cellphonedb_interactions:
        y = "cellphonedb_interactions"
    else:
        y = "interaction"
    
    # log (specificity +1)for plotting sized
    df["log_specificity"] = np.log10(df["specificity"]+1)
    
    # find and sort all connections
    connections = list(set(df["connection"]))
    connections.sort()

    
    
    plot = px.scatter(df, x="connection", y=y, 
        size="log_specificity", 
        color="log_score",
        opacity=0.8,
        height=height, width=width,
        size_max=10,
        template="plotly_white",
        render_mode="svg",
        title=f"Interactions with {filter_by} higher than {th}",
        color_continuous_scale=cmap,
        hover_data=["specificity"],
        category_orders = {"connection": connections})
    
    plot.update_layout(coloraxis_colorbar=dict(
        title="Log score",
        thicknessmode="pixels", thickness = 20,
        lenmode="pixels", len=200,
        yanchor="top", y=1
        )
    )
    return plot
    

# functions to retrieve dataframes of ligand and receptor
# information for further analysis
def get_ligand_df(G, color_map=False, corr_pval=True):
    """fetch all ligand information in a graph and return a DataFrame in long format"""
    import pandas as pd
    import numpy as np

    nodes = G.nodes(data=True)
    df = pd.DataFrame(columns=["ligand", "score", "node"])
    for node, data in G.nodes.items():
        temp_df = pd.DataFrame.from_dict(data["ligands_score"], orient="index", columns=["score"])
        temp_df["z_score"] = pd.DataFrame.from_dict(data["ligands_zscore"], orient="index", columns=["z-score"])
        if corr_pval:
            temp_df["pvalue"] = pd.DataFrame.from_dict(data["ligands_corr_pval"], orient="index", columns=["pvalue"])
        else:
             temp_df["pvalue"] = pd.DataFrame.from_dict(data["ligands_pval"], orient="index", columns=["pvalue"])
        temp_df["specificity"] = -np.log10(temp_df["pvalue"])
        temp_df["log_score"] = np.log10(temp_df["score"]+1)
        temp_df.rename_axis(index="ligand", inplace=True)
        temp_df["node"] = node
        temp_df.reset_index(inplace=True)
        df = df.append(temp_df, ignore_index=True,)
        
    color = dict(G.nodes(data="color"))
    for k, v in color.items():
        color[k] = "rgb" + str(tuple(v))
        
    if color_map == True:
        return df, color
    
    return df

def get_receptor_df(G, color_map=False, corr_pval=True):
    """fetch all ligand information in a graph and return a DataFrame in long format"""
    import pandas as pd
    import numpy as np

    nodes = G.nodes(data=True)
    df = pd.DataFrame(columns=["receptor", "score", "node"])
    for node, data in G.nodes.items():
        temp_df = pd.DataFrame.from_dict(data["receptors_score"], orient="index", columns=["score"])
        temp_df["z_score"] = pd.DataFrame.from_dict(data["receptors_zscore"], orient="index", columns=["z-score"])
        if corr_pval:
            temp_df["pvalue"] = pd.DataFrame.from_dict(data["receptors_corr_pval"], orient="index", columns=["pvalue"])
        else:
            temp_df["pvalue"] = pd.DataFrame.from_dict(data["receptors_pval"], orient="index", columns=["pvalue"])
        temp_df["specificity"] = -np.log10(temp_df["pvalue"])
        temp_df["log_score"] = np.log10(temp_df["score"]+1)
        temp_df.rename_axis(index="receptor", inplace=True)
        temp_df["node"] = node
        temp_df.reset_index(inplace=True)
        df = df.append(temp_df, ignore_index=True,)
        
    color = dict(G.nodes(data="color"))
    for k, v in color.items():
        color[k] = "rgb" + str(tuple(v))
        
    if color_map == True:
        return df, color
    
    return df



version = "2020-5"  # this is currently imported in other modules, so set verions here
organism = "mmusculus"

# Clean original GTP tables-----------------------------------------------------------------------------------------


def import_GTP_tables():
    """Imports all GTP tables into a dictionary of tables."""

    import pandas as pd
    import pkg_resources
    print(f"fetching GTF version {version}")
    # pkg_resourses can be used to access the local files in the package, independent of where the package is installed.
    i_path = pkg_resources.resource_filename(
        __name__, f"data/GTP_tables/{version}/interactions.csv")
    l_path = pkg_resources.resource_filename(
        __name__, f"data/GTP_tables/{version}/ligands.csv")
    p_path = pkg_resources.resource_filename(
        __name__, f"data/GTP_tables/{version}/peptides.csv")
    t_path = pkg_resources.resource_filename(
        __name__, f"data/GTP_tables/{version}/targets_and_families.csv")

    # imports full tables from Guidtopharmacology database
    interactions = pd.read_csv(i_path)
    interactions.index.name = "interaction_id"
    ligands = pd.read_csv(l_path, index_col=0)
    peptides = pd.read_csv(p_path, index_col=0)
    targets = pd.read_csv(t_path, index_col=3)

    return {"interactions": interactions, "ligands": ligands, "peptides": peptides, "targets": targets}


# Cleans tha names of targets and ligands in the target and ligands tables
def remove_html_tags(text, verbouse=True):
    """Remove html tags from a string"""

    import html
    import re

    clean = re.compile('<.*?>')
    if type(text) is str:
        try:
            clean_text = html.unescape(re.sub(clean, '', text))
            return clean_text
        except:
            if verbouse:
                print("could not clean ", text)
            return text


def clean_table(table):
    """Clean Html from any table and return the cleaned table."""

    return table.applymap(
        lambda text: remove_html_tags(text, verbouse=False)
    )


# Use this function to clean any new database tables.
def clean_GTP_tables():
    """Cleans all GTP tables from html and saves them as as new csv files."""

    import pkg_resources

    data = import_GTP_tables()
    for key, table in data.items():
        data[key] = clean_table(table)
        data[key].to_csv(pkg_resources.resource_filename(
            __name__, (f"data/GTP_tables_clean/{version}/{key}.csv")
        ))


# Input and Output----------------------------------------------------------------------------------------------
def get_data(data):
    """loads the data as a pandas DataFrame. 

    available tables are:

    - ligands
    - targets
    - interactions
    - peptides

    returns: pandas DataFrame
    """
    import pkg_resources
    import pandas as pd
    path = pkg_resources.resource_filename(
        __name__, (f"data/GTP_tables_clean/{version}/{data}.csv"))
    df = pd.read_csv(path)
    return df


# Annotation support ------------------------------------------------------------------------------------
    # Searches for orthogonal genes given an organism for which the gene exist
    # and a target organism for which to search.
def find_orth_gene(gene, organism, target):
    """Find orthogonal gene via Gprofiler

    returns a list of gene(s)"""
    import pandas as pd
    from gprofiler import GProfiler
    gp = GProfiler()

    if organism == target:  # do not search if original gene is known
        genes = list(set([gene, ]))
    else:
        results = pd.DataFrame(
            gp.orth(query=gene, organism=organism, target=target))
        results.dropna(subset=["name"], axis=0)
        genes = [gene for gene in results.name if gene != "N/A"]

    return genes


def get_peptide_ligands(organism=organism, save=True, verbouse=False):
    """Find genes encoding peptidergic ligands using gprofiler orthogonal gene search and
    the interactions table from GTP

    Saves updated peptide ligands to data/Gene_annotation/{organism}/peptide_ligands.csv
    Returns: pandas DataFrame
    """

    from scConnect.database import get_data
    import pandas as pd
    import numpy as np
    import pkg_resources
    import sys

    interactions = get_data("interactions")

    # Clean the dataframe removing duplicates of ligands from the same species
    dropped_dup = interactions.drop_duplicates(
        subset=["ligand", "ligand_species"])

    # All ligands need to have a gene and species attribute
    dropped_na = dropped_dup.dropna(
        axis=0, subset=["ligand_species", "ligand_gene_symbol"])

    # Subset dataframe to ligands, species and gene symbols
    gene_ligands = dropped_na[[
        "ligand", "ligand_species", "ligand_gene_symbol"]]

    # splits gene information from several species into induvidual records
    new_gene_list = []
    for row in gene_ligands.iterrows():
        ligand = row[1].ligand
        species = row[1].ligand_species.split("|")
        genes = row[1].ligand_gene_symbol.split("|")

        # Checks if two genes were reffered to.
        if len(species) < len(genes):
            genes = "|".join(genes)
            species_list = [ligand, str(species[0]), genes]
            new_gene_list.append(species_list)

        else:
            for i in np.arange(len(species)):
                species_list = [ligand, species[i], genes[i]]
                new_gene_list.append(species_list)

    gene_ligands_df = pd.DataFrame(new_gene_list, columns=[
        "ligand",
        "ligand_species",
        "ligand_gene_symbol"], dtype=str
    )
    # pivot gene symbols for each ligand and species.
    gene_pivot = gene_ligands_df.pivot(
        index="ligand", columns="ligand_species", values="ligand_gene_symbol")

    inferred_genes = list()

    # Add organism names. Key should match column names from GTP,
    # and values should match species names from gprofiler.
    # Note that the order of the dictionary matters (Py3.5>),
    # and if one gene is found ,the search complete.
    organism_dict = {
        "Mouse": "mmusculus",
        "Rat": "rnorvegicus",
        "Human": "hsapiens"
    }

    inferred_genes = list()
    for ligand in gene_pivot.index:
        genes = gene_pivot.loc[ligand]
        for name in organism_dict:
            if str(genes[name]) != "nan":
                gene = find_orth_gene(
                    gene=genes[name], organism=organism_dict[name], target=organism)
            else:
                continue

            if len(gene) > 0:
                comment = f"inferred from {name}"
                break
            else:
                gene = "nan"
                comment = "could not find gene"

        result = [ligand, gene, comment]

        inferred_genes.append(result)

        # add some output estimating % done
        percent = len(inferred_genes) / len(gene_pivot) * 100
        if verbouse:
            sys.stdout.write("\r" + str(np.around(percent, decimals=2)
                                        ) + "% done. \n Adding ligand: " + ligand)

    inferred_genes = pd.DataFrame(inferred_genes, columns=[
                                  "ligand", "gene", "comment"])

    inferred_genes = inferred_genes[inferred_genes.gene != "nan"]

    ligands = pd.DataFrame({"ligand": inferred_genes.ligand,
                            "ligand_type": "peptide",
                            "preprogene": inferred_genes.gene,
                            "synthesis": None,
                            "transport": None,
                            "reuptake": None,
                            "excluded": None,
                            "comment": inferred_genes.comment}
                           )

    if save == True:
        ligands.to_csv(pkg_resources.resource_filename(
            __name__, (f"data/Gene_annotation/{version}/{organism}/peptide_ligands.csv")))

    return ligands


def get_molecule_ligands(organism=organism, save=True):
    '''
    Creates a file with the molecule ligands

    Saves updated molecule ligands to data/Gene_annotation/molecule_ligands.csv
    returns: pandas DataFrame

    '''
    from scConnect.molecules import get_molecules
    import pkg_resources
    import pandas as pd

    mol = get_molecules()

    molecules = pd.DataFrame({
        "ligand": [molecule["ligand"] for molecule in mol],
        "ligand_type": [molecule["type"] for molecule in mol],
        "preprogene": [None for molecule in mol],
        "synthesis": [molecule["synthesis"] for molecule in mol],
        "transport": [molecule["transport"] for molecule in mol],
        "reuptake": [molecule["reuptake"] for molecule in mol],
        "excluded": [molecule["excluded"] for molecule in mol],
        "comment": ["Manually added genes" for molecule in mol]
    })

    drop_ligand = []  # detect if a critical gene category is not detected, then remove the ligand
    if organism != "mmusculus":
        types = ["synthesis", "transport", "reuptake", "excluded"]
        for ty in types:
            inferred_genes = {}
            for i, genes in enumerate(molecules[ty]):
                if genes != None:
                    inferred_genes[i] = list()
                    for gene in genes:
                        inferred_gene = find_orth_gene(
                            gene, organism="mmusculus", target=organism)
                        for gene in inferred_gene:
                            inferred_genes[i].append(gene)
                    if not len(inferred_genes[i]) > 0:
                        drop_ligand.append(i)
                else:
                    inferred_genes[i] = None
            molecules[ty] = inferred_genes.values()

    molecules.drop(index=list(set(drop_ligand)), inplace=True)
    molecules.reindex()

    if save is True:
        molecules.to_csv(pkg_resources.resource_filename(
            __name__, (f"data/Gene_annotation/{version}/{organism}/molecule_ligands.csv")))

    return molecules


def merge_ligand_lists(organism=organism, save=True):
    '''Returns a list with peptides and molecule ligands merged, 
    ready to be used with the package tools.

    saves updated ligands list to data/Gene_annotation/ligands.csv
    returns: pandas DataFrame
    '''
    import pandas as pd
    import pkg_resources

    molecules = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/{organism}/molecule_ligands.csv")), index_col=0)

    peptides = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/{organism}/peptide_ligands.csv")), index_col=0)

    ligands = peptides.append(molecules, ignore_index=True, sort=False)

    if save is True:
        ligands.to_csv(pkg_resources.resource_filename(
            __name__, (f"data/Gene_annotation/{version}/{organism}/ligands.csv")))
    return ligands


def get_receptors(organism=organism, receptor_types=["gpcr", "lgic"], save=True):
    '''
    returns a list of all receptors with known genes. 
    Finds ortholog genes for human, then mouse then rat.

    Many receptor types are available, defaults to only GPCRs and LGIC:
    enzyme                1259
    transporter            510
    gpcr                   416
    catalytic_receptor     307
    other_protein          228
    vgic                   145
    lgic                    86
    other_ic                54
    nhr                     49

    Saves updated receptor list to data/Gene_annotation/receptors.csv
    Returns: pandas DataFrame
    '''
    from scConnect.database import get_data
    import pandas as pd
    import pkg_resources

    receptors = get_data("targets")

    # Add all receptors to a list with correponding gene.
    # When two genes are found, these are added as a list.

    organism_dict = {  # decides in what order to look for orthological genes
        "HGNC symbol": "hsapiens",
        "MGI symbol": "mmusculus",
        "RGD symbol": "rnorvegicus"
    }

    receptor_list = list()
    for i, receptor in receptors.iterrows():
        if receptor["Type"] in receptor_types:  # select only receptors of specified types

            name = receptor["Target name"]
            family = receptor["Family name"]
            receptor_type = receptor["Type"]
            for species in organism_dict:
                if str(receptor[species]) != "nan":
                    genes = find_orth_gene(
                        gene=receptor[species], organism=organism_dict[species], target=organism)
                    break
                else:
                    genes = []
            if len(genes) > 0:
                receptor_list.append((name, family, receptor_type, genes))

    receptor_genes = pd.DataFrame(receptor_list, columns=[
                                  "receptor", "family", "type", "gene"])

    if save is True:
        receptor_genes.to_csv(pkg_resources.resource_filename(
            __name__, (f"data/Gene_annotation/{version}/{organism}/receptors.csv")))
    return receptor_genes


# build interaction table
def get_interactions(save=True):
    """build interaction table from GTP interaction df with only relevant data.

    Saves updated interaction table to data/Gene_annotation/interactions.csv.
    Returns: pandas DataFrame
    """
    import pkg_resources

    # Load data
    interactions = get_data("interactions")
    # remove unused receptors
    interactions = interactions.dropna(axis=0, subset=["target"])
    # Make a multiindex for quick indexing when finding interactions
    interactions.index = [interactions.ligand, interactions.target]

    # Keep useful columns and remove nan values for targets
    interactions = interactions[['endogenous',
                                 'ligand_species', 'target_species', 'pubmed_id', 'action']]

    # use ";" to seperate values due to "," being used in ligand names
    if save is True:
        interactions.to_csv(pkg_resources.resource_filename(
            __name__, (f"data/Gene_annotation/{version}/interactions.csv")), sep=";")

    return interactions


def setup_database(
    organism=organism,
    receptor_types=[
        "enzyme", "transporter", "gpcr", "catalytic_receptor",
        "other_protein", "vgic", "lgic", "other_ic", "nhr"]):
    '''Use this function to recalculate all
     ligands and receptor gene lists based on
     GTP tables located under data/GTP_tables.
     Manually create a folder for new species
     before running this function under /data/Gene_annotation.

     i.e. /data/Gene_annotation/mmusculus -- for mouse


     .. note:: 
        Download new `Guide to phalmacology data`_ and save the tables to data/GTP_tables.

    .. _Guide to phalmacology data: https://www.guidetopharmacology.org/download.jsp

    Many receptor types are available, defaults to only GPCRs and LGIC:
    enzyme                1259
    transporter            510
    gpcr                   416
    catalytic_receptor     307
    other_protein          228
    vgic                   145
    lgic                    86
    other_ic                54
    nhr                     49
    '''
    import scConnect as cn

    # Clean the new GTP tables from HTML
    cn.database.clean_GTP_tables()

    # Compute ligand genes from known genes and from manually added ligands
    # found in molecules.py
    print("getting peptide ligands...")
    pep = cn.database.get_peptide_ligands(organism=organism)
    print(f'found {pep.shape[0]} peptide ligands')

    print("getting molecule ligands...")
    mol = cn.database.get_molecule_ligands(organism=organism)
    print(f'found {mol.shape[0]} molecular ligands')

    print("merging ligands...")
    cn.database.merge_ligand_lists(organism=organism)

    print("getting receptors...")
    rec = cn.database.get_receptors(
        organism=organism, receptor_types=receptor_types)
    print(f'found {rec.shape[0]} receptors')
    print(rec.type.value_counts())

    print("getting interactions...")
    cn.database.get_interactions()

    print("Database build complete")



from setuptools import setup, find_packages
import glob

files = glob.glob("scConnect/**/*.*", recursive=True) # add all files under scConnect
files = [path.replace("\\", "/") for path in files] # Change from windows style to requred style for setuptools
files = [path.replace("scConnect/", "") for path in files]

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
   name='scConnect',
   version='1.0.3',
   description='scConnect integrate gene expression profiles in scRNA-seq datasets with ligand and receptor interaction information from Guide to pharmacology to construct a graph containing all putative interaction between cell types in the dataset.',
   long_description=long_description,
   author='Jon E.T. Jakobsson',
   author_email='jon.jakobsson@neuro.uu.se',
   Project_URL={"GitHub": "https://github.com/JonETJakobsson/scConnect",
                "Documentation": "https://scconnect.readthedocs.io/en/latest/"},
   license='License :: OSI Approved :: MIT License',
   key_words=['scRNA-seq, connectome, ligands, receptors, interactions, sequencing'],
   packages=["scConnect"],
   package_data={
      "scConnect": files},
   
   install_requires=[#external packages as dependencies
      'scanpy>=1.4',
      'pandas>=1.0',
      'numpy',
      'networkx',
      'dash',
      'dash-cytoscape',
      'jupyter-dash',
      'jupyter-server-proxy',
      'ipywidgets',
      'holoviews',
      'notebook',
      'xlrd'],
   Requires_Python='>=3.6', #utilize ordered dictionaries
   
)



import scConnect as cn
version = cn.database.version
# Module for different gene calling strategies
# Local functions


def get_adata_df(adata, transformation, layer, use_raw):
    """return a pandas DataFrame with gene expression counts and index and column lables.
    
    transformation: str or function, specify which transformation the data has been 
    subjected to so that count or UMI values can be calculated. By default, assumes that the data has been normalized using logp1.
    log1p back-transforms with expm1.
    Use a lambda call or provide a function that transforms the data back to the scale of gene counts. 
    It is fine to leave size factor normalization and linear scalings, but the data should not be logaritmised.
    layer: provide layer name to get specific layers.
    use_raw: weather to provide the stored raw data, or current adata.X matrix

    Returns: pandas DataFrame with values from specified layer
    """
    import pandas as pd
    from numpy import expm1

    # Get requested data
    if use_raw: # use raw if this is disired
        data = adata.raw.X.T
    elif layer != None: # else get the layer specified
        try:
            data = adata.layers[layer].T
        except: "layer not found"
    else: # if nothing is stated, provide current adata.X
        data = adata.X.T

    # transform data to counts
    if transformation == "log1p":
        data = expm1(data)

    if callable(transformation):
        print("using provided function")
        data = transformation(data)

    idx = adata.var_names
    col = adata.obs_names

    df = pd.DataFrame(data, index=idx, columns=col)
    return df


# Basic gene calling algorithms
def meanThreshold(adata, groupby, threshold, return_df=False, layer=None, use_raw=False, transformation="log1p"):
    """Binarize gene expression for groups aggregated by mean.

    Returns: adata object with updated uns.gene_call
    """
    from sklearn.preprocessing import binarize
    import pandas as pd
    df = get_adata_df(adata, layer=layer, use_raw=use_raw, transformation=transformation)
    result = df.groupby(by=adata.obs[groupby], axis=1).mean()
    binarize(result, threshold=threshold, copy=False)

    if return_df is True:
        return result
    else:
        adata.uns.update({"gene_call": result})
        return adata


def medianThreshold(adata, groupby, threshold, return_df=False, layer=None):
    """Binarize gene expression for groups aggregated by median.

    Returns: adata object with updated uns.gene_call
    """
    
    from sklearn.preprocessing import binarize
    import pandas as pd
    df = get_adata_df(adata, layer)
    result = df.groupby(by=adata.obs[groupby], axis=1).median()
    binarize(result, threshold=threshold, copy=False)

    if return_df is True:
        return result
    else:
        adata.uns.update({"gene_call": result})
        return adata


def meanExpression(adata, groupby, return_df=False, layer=None, use_raw=False, normalization=False, transformation="log1p"):
    """return gene expression for groups aggregated by mean.

    .. warning::
        Not compatible with certain molecular ligands as values are not between 0 and 1.
        Exclusion genes are weighted as:
        
        .. math:: Ligand or receptor call = Geom(Synthesis, Transporter, Reuptake) * (1 - Exclusion)
        
        Hence gene expression levels close to 1 contribute to 
        exlude the ligand and expression levels close to 0 leaves
        the ligand call untouched. 

    use_raw: Boolian, True if using the raw data of the adata structure. else False

    Normalization: Boolian, Normalize the gene expression over the different groups for each gene using L2 normalization.
    Note that this will lead to highly inflated values for lowly expressed genes, and under estimation of highly expressed genes.
    For experimental use only.

    Returns: adata object with updated uns.gene_call
    """

    import pandas as pd
    import numpy as np
    from sklearn.preprocessing import normalize

    df = get_adata_df(adata=adata, layer=layer, use_raw=use_raw, transformation=transformation)

    result = df.groupby(by=adata.obs[groupby], axis=1).mean()

    if normalization:
        df_norm = normalize(result, copy=True, axis=1, norm="l2")
        result = pd.DataFrame(df_norm, columns=result.columns, index=result.index)

    if return_df is True:
        return result
    else:
        adata.uns["gene_call"] = result.to_dict()
        return adata


def percentExpression(adata, groupby, return_df=False, layer=None, use_raw=False,):
    """Calculate percent of expressing cells in a cluster
    
    Returns: adata object with updated uns.gene_call
    """
    import pandas as pd
    from sklearn.preprocessing import binarize
    df = get_adata_df(adata, transformation=None, layer=layer, use_raw=use_raw)
    binarize(df, threshold=0.0, copy=False)
    result = df.groupby(by=adata.obs[groupby], axis=1).mean()

    if return_df is True:
        return result
    else:
        adata.uns.update({"gene_call": result})
        return adata


# Implementing betabinomial trinarization as seen in Zeisel et al. 2018, Cell

def betabinomialTrinarization(adata, groupby, pep=0.05, f=0.2, return_df=False, verbouse=True, layer=None):
    """Trinarize gene expression for groups.
    
    Function is adapted from `Zeisel 2018`__.

    __ https://www.sciencedirect.com/science/article/pii/S009286741830789X


    pep: posterior error probability, float (0-1)
    
    f: fraction, float (0-1)

    Returns: adata object with updated uns.gene_call
    """
    import pandas as pd
    from collections import Counter
    from scConnect.tools import printProgressBar

    df = get_adata_df(adata, layer)
    labels = adata.obs[groupby]
    clusters = dict(Counter(labels)).keys()
    gene_call = pd.DataFrame(columns=clusters)
    tot_genes = df.shape[0]
    called_gene = 0
    for gene, expression in df.iterrows():
        expression_by_label = betabinomial_trinarize_array(
            expression, labels, pep, f)
        gene_call.loc[gene] = expression_by_label
        called_gene += 1
        if verbouse:
            printProgressBar(called_gene, tot_genes,
                             prefix=f"Calling gene expression using betabinomial trinarization [Zeisel et al. 2018] gene: {called_gene} of {tot_genes}")
    if return_df is True:
        return gene_call
    else:
        adata.uns.update({"gene_call": gene_call})
        return adata


def betabinomial_trinarize_array(array, labels, pep, f, n_labels=None):
    import numpy as np
    from collections import Counter

    if n_labels is None:
        #n_labels = np.max(labels) + 1
        n_labels = len(set(labels))

    n_by_label = dict(Counter(labels))
    k_by_label = dict()

    for lbl in n_by_label.keys():
        try:
            k_by_label[lbl] = np.count_nonzero(
                array[np.where(labels == lbl)[0]])
        except KeyError:
            print(f"no cells from cluster {lbl}")
            k_by_label[lbl] = 0

    ns = [n for n in n_by_label.values()]
    ks = [k for k in k_by_label.values()]

    vfunc = np.vectorize(p_half)
    ps = vfunc(ks, ns, f)

    expr_by_label = np.zeros(n_labels) + 0.5
    expr_by_label[np.where(ps > (1 - pep))[0]] = 1
    expr_by_label[np.where(ps < pep)[0]] = 0

    return (expr_by_label)


def p_half(k, n, f):
    from math import exp, lgamma, log
    from scipy.special import beta, betainc, betaln
    import numpy as np

    # These are the prior hyperparameters beta(a,b)
    a = 1.5
    b = 2

    incb = betainc(a + k, b - k + n, f)

    if incb == 0:
        p = 1.0
    else:
        p = 1.0 - exp(log(incb) + betaln(a + k, b - k + n) +
                      lgamma(a + b + n) - lgamma(a + k) - lgamma(b - k + n))

    return p


#Calculate new mean values based on Zero inflation model
def ziMean(adata, groupby, organism="mmusculus"):
    """Calculates mean expression based on estimated dropouts given the mean expression of non zero cells.
    
    Using the decay coeficient calculated by the ZIFA algorithm, estimated dropput rates can 
    be calculated based on the mean expression of non zero counts. New mean expression values is then calculated
    from imputed counts.
    """

    import ZIFA.block_ZIFA as zf
    import scanpy as sc
    adata_copy = adata.copy()
    adata_copy = filter_genes(adata_copy, organism=organism)
    sc.pp.filter_genes(adata_copy, min_counts=1)
    countmatrix = adata_copy.X
    model, params = zf.fitModel(countmatrix, 2, singleSigma=True)
    dc = params["decay_coef"]
    df = get_adata_df(adata_copy)

    def ZImean(countmatrix):
        """Calculated new mean values given set decay coeficiant. Can be used with groupby.aggregate"""

        import numpy as np
        gene_call = dict()
        for gene, expression in countmatrix.iterrows():
            total_cells = len(expression)
            expressing_cells = len(expression[expression>0])
            if expressing_cells>0:
                non_zero_mean = np.mean(expression[expression>0])
                prob = np.exp(-dc*non_zero_mean**2)
                with_dropout = int(np.round(expressing_cells/(1-prob), decimals=0))
                if with_dropout > total_cells:
                    with_dropout = total_cells

                new_counts = [non_zero_mean]*with_dropout+[0]*(total_cells-with_dropout)
                new_mean = np.mean(new_counts)
            else:
                new_mean = 0
            gene_call[gene] = new_mean
    
        return gene_call

    
    gene_call = df.groupby(adata_copy.obs[groupby], axis=1).aggregate(ZImean)

    adata.uns.update({"gene_call": gene_call})
    return adata
# Plotting tools to investigate gene calling performance.


def geneParticipation(adata):
    """Plot histogram showing % of cluster genes are prescent in."""
    import numpy as np
    df = adata.uns["gene_call"]
    cluster_participation = df.aggregate(np.mean, axis=1)
    cluster_participation.plot(kind="hist", bins=20)


def filter_genes(adata, organism="mmusculus"):
    """Filter adata object to only have ligand and receptor genes.
    
    .. note ::
        Usefull to create ligand and receptor centric clustering and to speed up receptor and ligand calling.
    """
    import numpy as np
    import pandas as pd
    import pkg_resources

    ligands = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/{organism}/ligands.csv")))
    receptors = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/{organism}/receptors.csv")))

    def add_gene(gene, gene_list):
        try:
            gene = eval(str(gene))
        except (NameError, TypeError):
            pass
        if type(gene) == str:
            gene_list.append(gene)
            return gene_list
        elif type(gene) == list:
            for g in gene:
                gene_list.append(g)
            return gene_list
        else:
            return gene_list

    gene_list = list()
    ligands = ligands[["preprogene", "synthesis",
                       "transport", "reuptake", "excluded"]]
    for column in ligands.columns:
        for gene in ligands[column]:
            gene_list = add_gene(gene, gene_list)

    for gene in receptors["gene"]:
        gene_list = add_gene(gene, gene_list)

    gene_list = list(set(gene_list))
    filter_genes = [
        True if gene in gene_list else False for gene in adata.var_names]
    adata = adata[:, filter_genes]
    return adata

# Finding optimal settings using Glutamate and GABA segregation


def segregation_score(adata):
    """Calculate a score of segregation of gabaergic and glutamatergic cells"""
    import pandas as pd
    total_score = []
    df = pd.DataFrame(adata.uns["ligands"]).loc[["GABA", "L-glutamic acid"]]
    for cluster in df:
        gaba = df[cluster].loc["GABA"] > 0
        glut = df[cluster].loc["L-glutamic acid"] > 0

        if gaba and glut:  # cell population expresses both ligands
            total_score.append(-1)
        elif not gaba and not glut:  # cell population expresses no ligands
            total_score.append(-1)
        else:  # cell population expresses only one ligand
            total_score.append(1)

    # normalize on total number of cell populations
    return(sum(total_score)/len(total_score))


def plot_segregation(adata, save=False, filename=None):
    """Plot gabaergic and glutamaterig cell populations"""
    import holoviews as hv
    from holoviews import opts
    import pandas as pd
    hv.extension("matplotlib")

    df = pd.DataFrame(adata.uns["ligands"]).loc[[
        "GABA", "L-glutamic acid"]].stack().reset_index()
    df.columns = ["ligand", "cluster", "value"]
    df = df.sort_values(by="cluster", axis=0)

    opts.defaults(
        opts.Bars(
            stacked=True,
            xrotation=90,
            legend_position="right",
            ylabel="Ligand score"
        )
    )
    bars = hv.Bars(df, kdims=["cluster", "ligand"])
    if save is True:
        hv.save(bars, filename)
    return bars


def optimize_segregation(adata, groupby, method="meanTH", start=0, stop=1, steps=10, organism="mmusculus", fraction=0.8, iterations=10, save=False, filename=None, return_df=False, verbouse=False, transformation="log1p", use_raw=False):
    """Calculate segregation score for a range of thresholds using specified method.
    
    available methods:
    meanTH
    medianTH
    NBT

    The estimation will be iterated over selected number of times, and an average segregation will 
    be plotted. error area is 99% CI.

    Returns: Seaborn lineplot of score over threshold

    Save file using save=True and a file name with .png or.svg extention
    """
    from collections import OrderedDict
    import scanpy as sc
    from random import randint, sample
    import seaborn as sns
    import pandas as pd
    import numpy as np
    import scConnect as cn
    from scConnect.tools import printProgressBar

    # filter out only genes nessesary to calculate gaba and glutamate presence
    mol_ligands = cn.database.get_molecule_ligands(organism=organism)
    seg_filter = [True if ligand in ["GABA", "L-glutamic acid"] else False for ligand in mol_ligands["ligand"]]
    mol_ligands = mol_ligands[seg_filter][["synthesis","transport","reuptake"]]
    gene_list = list()
    for genes in mol_ligands.values.flatten():
        if genes is not None:
            for gene in genes:
                    gene_list.append(gene)

    adata = adata[:, gene_list]
    
    th_score = list()
    space = np.linspace(start, stop, steps)
    tot_steps = len(space)*iterations
    cur_step = 1
    obs = adata.obs_names

    for th in space:
        for iteration in np.arange(0,iterations):

            # Create random subset of dataset for distribution calculation
            obs_s = sample(set(obs), int(len(obs)*fraction))
            temp = adata[obs_s]

            #Store method calls with relevant parameters
            methods = {
                "NBT": "betabinomialTrinarization(temp, groupby=groupby, f=th, verbouse=False)",
                "medianTH": "medianThreshold(temp, groupby=groupby, threshold=th, layer=None, use_raw=use_raw, transformation=transformation)",
                "meanTH": "meanThreshold(temp, groupby=groupby, threshold=th, layer=None, use_raw=use_raw, transformation=transformation)"
                }

            #Runs selected method with relevant local variabls
            temp = eval(methods[method])
            temp = cn.connect.ligands(temp, organism=organism, select_ligands=[
                                    "GABA", "L-glutamic acid"])
            score = segregation_score(temp)
            th_score.append((iteration, th, score))
            if verbouse:
                printProgressBar(cur_step, tot_steps,
                                prefix=f"Finding optimized threshold")
            cur_step += 1

    # store data in a pandas dataframe
    df = pd.DataFrame(data=th_score, columns=["Iteration", "Threshold", "Segregation"])

    # group by threshold to calculate best score and best threshold
    g_df = df.groupby(by="Threshold")["Segregation"].mean()

    # idxmax will return the first index with max score, which is what we want.
    max_segregation = g_df[g_df.idxmax()]
    best_th = g_df.idxmax()
    print(f"Maximun segregation of {max_segregation} was achived at threshold {best_th}")


    if return_df:
        return df

    f = sns.lineplot(x=df["Threshold"], y=df["Segregation"], ci=99, estimator="mean")
    f.axvline(best_th, ls="--", color="red")

    if save is True:
        fs = f.get_figure()
        fs.savefig(filename)
    return f


# Module deals with creation of ligand and receptor scores, and creation of scConnect tables etc.
import scConnect as cn
import scanpy as sc

version = cn.database.version
organism = cn.database.organism
# Scoring logic for ligands


def ligandScore(ligand, genes):
    """calculate ligand score for given ligand and gene set"""

    from scipy.stats.mstats import gmean
    import numpy as np

    if ligand.ligand_type == "peptide" and isinstance(ligand.preprogene, str):
        # check if multiple genes needs to be accounted for
        if isinstance(eval(ligand.preprogene), list):
            ligand_genes = list()
            for gene in eval(ligand.preprogene):
                try:
                    ligand_genes.append(genes[gene])
                except KeyError:
                    #print(f"{gene} not found")
                    ligand_genes.append(0.0)
            # use max, as there might be many orthologs genes for one original 
            # gene and not all have to be expressed
            try:
                ligand_score = max(ligand_genes)
            except ValueError:
                print(f"something is wrong with the list {ligand_genes}")
                ligand_score = 0.0

            return ligand_score


    elif ligand.ligand_type == "molecule":
        synthesis = ligand.synthesis
        transport = ligand.transport
        reuptake = ligand.reuptake
        excluded = ligand.excluded

        # get geometric mean of synthesis genes (all need to be present)
        if not isinstance(synthesis, str):
            # If no genes are needed, synthesis is set to nan
            synthesis = np.nan
        else:
            synthesis_expression = list()
            for gene in eval(synthesis):
                try:
                    synthesis_expression.append(genes[gene])
                except KeyError:
                    # If gene was not found append 0
                    #print(f"{gene} not found")
                    synthesis_expression.append(0.0)
            synthesis = gmean(synthesis_expression)

        # get maximum of vesicle transporters (only one is needed for molecule transport)
        if not isinstance(transport, str):
            # If no specific genes are needed, set transport to nan
            transport = np.nan
        else:
            transport_expression = list()
            for gene in eval(transport):
                try:
                    transport_expression.append(genes[gene])
                except KeyError:
                    # If gene was not found append 0
                    #print(f"{gene} not found")
                    transport_expression.append(0.0)
            transport = max(transport_expression)

        # Get maximum of reuptake genes (only one is needed)
        if not isinstance(reuptake, str):
            # If no specific genes are needed, set reuptake to nan
            reuptake = np.nan
        else:
            reuptake_expression = list()
            for gene in eval(reuptake):
                try:
                    reuptake_expression.append(genes[gene])
                except KeyError:
                    # If gene was not found append 0
                    #print(f"{gene} not found")
                    reuptake_expression.append(0.0)
            reuptake = max(reuptake_expression)

        # get maximum among exluding genes where any gene expression divert to other ligands
        if not isinstance(excluded, str):
            # If no specific genes are needed, set excluded to 0
            excluded = 0
        else:
            excluded_expression = list()
            for gene in eval(excluded):
                try:
                    excluded_expression.append(genes[gene])
                except KeyError:
                    # If gene was not found append 0
                    #print(f"{gene} not found")
                    excluded_expression.append(0.0)
            excluded = max(excluded_expression)

        # return geometric mean of synthesis, transport and reuptake multipled exclusion
        promoting_factor = gmean(([x for x in [synthesis, transport, reuptake] if str(x) != "nan"])) # genes driving ligand production, remove nan values
        
        if str(promoting_factor) == "nan": # capture cases where no promoting genes were present
            print(f"no promoting genes detected for {ligand.ligand}")
            return 0.0  # exit before running exclusion calculation
            
        ligand_score = promoting_factor - excluded # correct ligand expression based on the exclusion factor
        if ligand_score < 0: # ligand score should be 0 or positive
                ligand_score = 0.0
        return ligand_score

    # If genes are missing from ligand gene list
    else:
        print("Big error! ligand type is not defined!")
        return 0.0


def ligands(adata, organism=organism, select_ligands=None):
    """return a dataframe with ligand scores for each cluster.

    .. note::
        Needs a gene call dataframe under adata.uns.gene_call. 
        Use scConnect.genecall to create such dataframe

    organism defaults to mouse, to use genes for other organism select this here.

    use select_ligands to only asses given ligands 
    (used by optimize_segregation to only check for gaba and glutamate)

    Returns: Dict of ligand call for each cluster.
    """
    import scConnect as cn
    import pkg_resources
    import pandas as pd

    ligands = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/{organism}/ligands.csv")))
    if isinstance(select_ligands, list):
        select = [True if ligand in select_ligands else False for ligand in ligands.ligand]
        ligands = ligands[select]
    ligand_df = pd.DataFrame(index=ligands.ligand)
    
    for cluster, genes in adata.uns["gene_call"].items():
        cluster_scores = list()
        for ligand_data in ligands.iterrows():
            ligand = ligand_data[1]
            # fetch ligand score for specific ligand and gene set
            ligand_score = ligandScore(ligand, genes)
            cluster_scores.append(ligand_score)
        ligand_df[cluster] = cluster_scores

    adata.uns["ligands"] = ligand_df.to_dict()
    return adata


# Scoring logic for receptors
def receptorScore(receptor, genes):
    """calculate receptor score given receptor and gene set"""

    from scipy.stats.mstats import gmean

    gene_expression = list()
    for gene in eval(receptor.gene):
        try:
            gene_expression.append(genes[gene])
        except KeyError:
            # If gene was not found append 0
            #print(f"{gene} not found")
            gene_expression.append(0.0)
    # use max, as several genes might be found during ortholog search, 
    # not all might bee needed to create the receptor
    gene_expression = max(gene_expression)
    return gene_expression


def receptors(adata, organism=organism):
    """return a dataframe with receptor scores for each cluster.

    .. note::
        Needs a gene call dataframe under adata.uns.gene_call. 
        Use scConnect.genecall to create such dataframe.

    Returns: Dict of receptor call for each cluster.
    """
    import scConnect as cn
    import pkg_resources
    import pandas as pd

    receptors = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/{organism}/receptors.csv")))
    receptor_df = pd.DataFrame(index=receptors.receptor)

    for cluster, genes in adata.uns["gene_call"].items():
        cluster_scores = list()
        for receptor_data in receptors.iterrows():
            receptor = receptor_data[1]
            # fetch ligand score for specific ligand and gene set
            receptor_score = receptorScore(receptor, genes)
            cluster_scores.append(receptor_score)
        receptor_df[cluster] = cluster_scores
    adata.uns["receptors"] = receptor_df.to_dict()
    return adata


# Interaction logic

def interactions(emitter, target, self_reference=True, organism=organism, corr_pval=True):
    """return an edge list of interactions between clusters.
    If all connections are of interest, use the same data source for
    emitter and target.

    .. note::
        self_reference is only valid when emitter == target.

    .. note::
        edge_list is returned as a list, and not in a adata object. 
        This is since multiple adata objects can be passed in to the
        function, and whould lead to ambiguity of which object to append the edge_list to.
    
    Returns: List of edges between given emmitor and target clusters.
    """
    import pkg_resources
    import pandas as pd
    from itertools import product
    from scConnect.tools import printProgressBar

    interactions = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/interactions.csv")), index_col=[0, 1], sep=";")

    interactions.sort_index(axis="index", inplace=True)
    # Create a set of all possible index combinations.
    # This is used to test if ligand receptor combination is present.
    interaction_set = set(interactions.index)

    # An edge list should contain u, v and d,
    # where u is input node, v is output node
    # and d is a dictionary with edge attributes.
    edge_list = list()

    # get all clusters
    # NOTE: if the same cluster name is used in emitter and target datasets, they are
    # assumed to be the same cluster. Give your clusters uniqe names between your datasets.
    try:
        emitter_clusters = pd.DataFrame(emitter.uns["ligands"]).columns
        target_clusters = pd.DataFrame(target.uns["ligands"]).columns
    except KeyError:
        print(
            f"Please run connect.ligands() and connect.receptors() on your datasets first")
        return

    # Calculate total number of cluster combinations for the progress bar
    if self_reference is True:
        total_comb = len(list(product(emitter_clusters, target_clusters)))
    else:
        total_comb = len([(e, t) for (e, t) in product(
            emitter_clusters, target_clusters) if e != t])

    ligands = pd.DataFrame(emitter.uns["ligands"])
    receptors = pd.DataFrame(target.uns["receptors"])

    # load extra ligand and receptor statistics
    ligands_zscore = pd.DataFrame(emitter.uns["ligands_zscore"])
    receptors_zscore = pd.DataFrame(target.uns["receptors_zscore"])
    if corr_pval:
        ligands_pval = pd.DataFrame(emitter.uns["ligands_corr_pval"])
        receptors_pval = pd.DataFrame(target.uns["receptors_corr_pval"])
    else:
        ligands_pval = pd.DataFrame(emitter.uns["ligands_pval"])
        receptors_pval = pd.DataFrame(target.uns["receptors_pval"])

    # Fetch receptor and ligand information
    receptor_info = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/{organism}/receptors.csv")), index_col=1)
    receptor_info = receptor_info[["family", "gene"]]

    ligand_info = pd.read_csv(pkg_resources.resource_filename(
        __name__, (f"data/Gene_annotation/{version}/{organism}/ligands.csv")), index_col=1)
    ligand_info = ligand_info[["ligand_type", "comment"]]

    # Nested for-loop to get all combinations of
    # interactions between clusters.
    comb_tried = 0
    for emitter_cluster in emitter_clusters:
        for target_cluster in target_clusters:
            # Are we interested in self referencing information?
            # I leave that up to the user
            if emitter_cluster != target_cluster or self_reference == True:
                # Get only ligands and receptors expressed by the clusters
                # (speeds up itterative functions later)
                emitter_ligands = ligands[emitter_cluster][ligands[emitter_cluster] > 0]
                target_receptors = receptors[target_cluster][receptors[target_cluster] > 0]

                connections = get_connections(
                    emitter_ligands, 
                    target_receptors, 
                    interactions, 
                    interaction_set, 
                    receptor_info, 
                    ligand_info,
                    emitter_cluster,
                    target_cluster,
                    ligands_zscore,
                    ligands_pval,
                    receptors_zscore,
                    receptors_pval)

                if len(connections) > 0:
                    for connection in connections:
                        edge_list.append(connection)

                # Add the progress bar
                comb_tried += 1
                printProgressBar(
                    comb_tried, total_comb, prefix=f"finding connections between {len(emitter_clusters)} emitter clusters and {len(target_clusters)} target clusters")

    return edge_list


# get all connections based on Ligands and receptors, and provide score for interactions
# Also provide meta data as a dictionary for interaction
def scale(value, from_range=(0, 1), to_range=(10E-100, 1)): # mitagate log with 0
        value = to_range[0] + (to_range[1] - to_range[0]) * (value -from_range[0]) / (to_range[1] - to_range[0])
        return value

def get_connections(
    ligands,
    receptors, 
    interactions, 
    interaction_set, 
    receptor_info, 
    ligand_info,
    emitter_cluster,
    target_cluster, 
    ligands_zscore,
    ligands_pval,
    receptors_zscore,
    receptors_pval):
    """finds connections between ligands and receptors 
    and return a score and metadata for each interaction"""

    from scipy.stats.mstats import gmean
    import numpy as np
    # shorten the list of interactions to only contain relevant ligands.
    # This should speed up the algorithm
    ligand_filter = [True if ligand in ligands.keys() else False for ligand in interactions.index.get_level_values(0)]
    interactions = interactions.loc[ligand_filter]

    

    def interaction_specificity(l, r): # used to calculate interaction specificity score
        sig = -np.log10((l+r)/2)
        return sig

    connections = list()
    for ligand, l_score in ligands.iteritems():
        for receptor, r_score in receptors.iteritems():
            if (ligand, receptor) in interaction_set:
                interaction = interactions.loc[ligand, receptor]
                score = float(gmean((l_score, r_score)))
                ligand_pval = float(ligands_pval[emitter_cluster][ligand])
                receptor_pval = float(receptors_pval[target_cluster][receptor])
                specificity = float(interaction_specificity(ligand_pval, receptor_pval))
                log_score = float(np.log10(score + 1))
                importance = specificity * log_score

                connections.append((ligands.name, receptors.name, {
                    "score": float(score),
                    "log_score": log_score, # From here on, all values are +1ed and logaritmized with base of 10. # From here on, all values are +1ed and logaritmized with base of 10.
                    "ligand": str(ligand),
                    "ligand_zscore": float(ligands_zscore[emitter_cluster][ligand]),
                    "ligand_pval": ligand_pval,
                    "receptor": str(receptor),
                    "receptor_zscore": float(receptors_zscore[target_cluster][receptor]),
                    "receptor_pval": receptor_pval,
                    "interaction": f"{ligand} --> {receptor}",
                    "specificity": specificity,
                    "importance": importance,
                    "endogenous": f"{list(interaction.endogenous)}",
                    "action": f"{list(interaction.action)}",
                    "ligandspecies": f"{list(interaction.ligand_species)}",
                    "receptorspecies": f"{list(interaction.target_species)}",
                    "pubmedid": f"{list(interaction.pubmed_id)[:5]}",
                    "receptorfamily": str(receptor_info.loc[receptor]["family"]),
                    "receptorgene": str(receptor_info.loc[receptor]["gene"]),
                    "ligandtype": str(ligand_info.loc[ligand]["ligand_type"]),
                    "ligandcomment": str(ligand_info.loc[ligand]["comment"])}))
    return connections


def nodes(adatas):
    """
    Returns an list of nodes, attributes dictionary tuples. 
    Each tuple represent one node with an attribute dictionary: 
    *(cluster, dict(receptors: dict(receptor:score), ligands: dict(ligand:score) ))*
    
    """
    if not isinstance(adatas, list):
        adatas = [adatas, ]

    nodes = []

    for i, adata in enumerate(adatas):
        print(f"precessing adata #{i+1}")
        ligands_score = adata.uns["ligands"]
        ligands_zscore = adata.uns["ligands_zscore"]
        ligands_pval = adata.uns["ligands_pval"]
        ligands_corr_pval = adata.uns["ligands_corr_pval"]
        receptors_score = adata.uns["receptors"]
        receptors_zscore = adata.uns["receptors_zscore"]
        receptors_pval = adata.uns["receptors_pval"]
        receptors_corr_pval = adata.uns["receptors_corr_pval"]
        genes = adata.uns["gene_call"]
        clusters = ligands_score.keys()

        # Filter out ligands with positive score (remove non expressing ligands and receptors)
        for cluster in clusters:
            print(f"processing cluster {cluster}")
            cluster_ligands_score = {k: v for k,
                               v in ligands_score[cluster].items() if v > 0}
            cluster_receptors_score = {k: v for k,
                                 v in receptors_score[cluster].items() if v > 0}

            # Add all information to the node dicionary
            node = (cluster, {
                "ligands_score": cluster_ligands_score,
                "ligands_zscore": ligands_zscore[cluster],
                "ligands_pval": ligands_pval[cluster],
                "ligands_corr_pval": ligands_corr_pval[cluster],
                "receptors_score": cluster_receptors_score,
                "receptors_zscore": receptors_zscore[cluster],
                "receptors_pval": receptors_pval[cluster],
                "receptors_corr_pval": receptors_corr_pval[cluster],
                "genes": genes[cluster]})
            nodes.append(node)

    return nodes

# Statistic inference of ligand and receptor scores
# Here we shuffel the group annotations many times, calculate ligand and receptor scores
# and find the mean and standard deviation for each ligand/receptor score for each gorup.
# We can then calculate the z-score of the true ligand/receptor score, p-values and corrected p-values
# Data an be used to detect group specific expression of ligands and receptors.

def _ligand_receptor_call(adata, groupby, organism, transformation, return_df = True):
    import pandas as pd
    adata = cn.genecall.meanExpression(adata, groupby=groupby, normalization=False, use_raw=False, transformation=transformation)
    adata = cn.connect.ligands(adata, organism=organism)
    adata = cn.connect.receptors(adata, organism=organism)
    
    ligands = pd.DataFrame(adata.uns["ligands"])
    receptors = pd.DataFrame(adata.uns["receptors"])
    if return_df:
        return ligands, receptors

def _values_df(dfs):
    values_df = dfs[0].copy()
    
    for i in range(values_df.shape[0]):
        for j in range(values_df.shape[1]):
            values = list()
            for df in range(len(dfs)):
                values.append(dfs[df].iloc[i,j])
            values_df.iloc[i,j] = str(values)
    return values_df

            
def _mean_df(df):
    import numpy as np
    mean_df = df.copy()
    for i in range(df.shape[0]):
        for j in range(df.shape[1]):
            mean_df.iloc[i,j] = np.mean(eval(df.iloc[i,j]))
    return mean_df
        
def _std_df(df):
    import numpy as np
    std_df = df.copy()
    for i in range(df.shape[0]):
        for j in range(df.shape[1]):
            std_df.iloc[i,j] = np.std(eval(df.iloc[i,j]))
    return std_df
        

def _score_pv_df(mean, std, value, emperical, values, merge_dist):
    """Calculate z-scores and p-values for ligand and receptor 
    calls compared to random group designation
    
    returns: score_df and pval_df"""
    import numpy as np
    from scipy import stats

    assert (mean.shape == std.shape == value.shape), "dataframes are not of the same size, rerun ligand and receptor call"

    score_df = mean.copy()
    pval_df = mean.copy()
    warning = False # warning flag for if mean or std is 0 (mening no values were ever sampled to that group)
    faults = 0

    for i in range(score_df.shape[0]): # for each ligand and receptor
        if merge_dist == True:
            dist = list()
            for j in range(score_df.shape[1]):
                for val in eval(values.iloc[i,j]):
                    dist.append(val)
                
        for j in range(score_df.shape[1]): # for each celltype
            v = value.iloc[i,j]
            s = std.iloc[i,j]
            m = mean.iloc[i,j]
            if merge_dist == False:
                dist = eval(values.iloc[i,j])
        
            if s == 0: # sampeling never managed to include this ligand or receptor for this group
                z_score = 0.0
                pval = 1
                warning = True
                faults += 1
            else:
                z_score = (v-m)/s
                #pval =  float(stats.norm.sf(abs(z_score))*2) # Two tailed p-value
                pval =  float(stats.norm.sf(z_score)) # one tailed p-value

            if emperical == True: 
                # Calculate exact permutation pvalues emperically from the collected distribution
                # method from https://www.degruyter.com/view/journals/sagmb/9/1/article-sagmb.2010.9.1.1585.xml.xml
                # permutation without replacement (use full sequence)
                b = sum(dist > v)
                m = len(dist)
                pval = (b+1)/(m+1)

            
            score_df.iloc[i,j] = z_score
            pval_df.iloc[i,j] = pval
            
    if warning:
        total = score_df.shape[0] * score_df.shape[1]
        print(f"{faults/total*100} % of group metrices were 0. increase n to reduce this number")
    
    return score_df, pval_df

def _corrected_pvalue(pvalues, method="fdr_bh", scale_pval=False):
    """correct a dataframe of p-values to a dataframe of corrected p-values.
    
    Supports many different methods:
    bonferroni : one-step correction
    sidak : one-step correctio
    holm-sidak : step down method using Sidak adjustment
    holm : step-down method using Bonferroni adjustment
    simes-hochberg : step-up method (independent
    hommel : closed method based on Simes tests (non-negative)
    fdr_bh : Benjamini/Hochberg (non-negative)
    fdr_by : Benjamini/Yekutieli (negative)
    fdr_tsbh : two stage fdr correction (non-negative)
    fdr_tsbky : two stage fdr correction (non-negative)
    
    defaults to fdr_bh
    
    returns a pandas dataframe
    """
    import statsmodels.stats.multitest as mt
    import pandas as pd
    
    p_flat = pvalues.to_numpy().flatten()
    corr_p = mt.multipletests(p_flat, method=method)[1]
    
    corr_p = corr_p.reshape(pvalues.shape)
    corr_pval = pd.DataFrame(corr_p, columns=pvalues.columns, index=pvalues.index)
    
    # scale p values to remove abloslute 0 calls
    if scale_pval:
        corr_pval = scale(corr_pval)

    return corr_pval
    
def specificity(adata, n, groupby, organism=organism, return_values=False, transformation="log1p", emperical=True, merge_dist=False):
    """calculate statistics for the ligands and receptor scores.
    
    Compare the group ligand and receptor scores to the mean score of 
    that group after n number of permutations
    
    if emperical is True (default), calculates p-values emperically given the collected random distribution.
    p = (b+1)/(m+1) 
    where b is the number of permutated values higher than the observed
    and m is the number of permutations used (set this by the argument n)"""

    from random import shuffle
    import pandas as pd
    from scConnect.tools import printProgressBar
    _adata = adata.copy()
    groups = list(_adata.obs[groupby])
    
    ligand_dfs = list()
    receptor_dfs = list()
    
    # variable to store the setting in, can be used when saving the specificity
    settings = dict(
        groupby = groupby,
        permutations =  n,
        transformation = transformation,
        emperical = emperical,
        merge_dist = merge_dist
    )

    # Run normal ligand and receptor call without shuffel on original adata
    _ligand_receptor_call(adata, groupby=groupby, organism=organism, transformation=transformation, return_df=False)
    # shuffel group annotations n times and fetch ligand and receptor dataframes
    for i in range(n):
        printProgressBar(i+1, n, prefix=f"Shuffeling dataframe {i+1} out of {n}")
        shuffle(groups)
        _adata.obs[groupby] = groups
        ligand, receptor = _ligand_receptor_call(_adata, groupby=groupby, organism=organism, transformation=transformation)
        ligand_dfs.append(ligand)
        receptor_dfs.append(receptor)
    
    # Merge all dataframes to one datafram (with list of values for each element)
    ligand_values = _values_df(ligand_dfs)
    receptor_values = _values_df(receptor_dfs)
    
    # Calculate the mean values of the list in each element
    print("Calculating means...")
    ligand_mean = _mean_df(ligand_values)
    receptor_mean = _mean_df(receptor_values)
    
    # Calculate the standard deviation of the list in each element
    print("Calculating standard deviations...")
    ligand_std = _std_df(ligand_values)
    receptor_std = _std_df(receptor_values)
    
    # Calculate Z-scores, p-values and corrected p-values
    print("Calculating Z-score, p-values and corrected p-values...")
    ligand_value = pd.DataFrame(adata.uns["ligands"])
    ligand_score , ligand_pval = _score_pv_df(ligand_mean, ligand_std, ligand_value, emperical, ligand_values, merge_dist=merge_dist)
    ligand_corr_pval = _corrected_pvalue(ligand_pval, scale_pval=not emperical)
    
    receptor_value = pd.DataFrame(adata.uns["receptors"])
    receptor_score , receptor_pval = _score_pv_df(receptor_mean, receptor_std, receptor_value, emperical, receptor_values, merge_dist=merge_dist)
    receptor_corr_pval = _corrected_pvalue(receptor_pval, scale_pval=not emperical)
    
    adata.uns.update({"ligands_zscore": ligand_score.to_dict()})
    adata.uns.update({"receptors_zscore": receptor_score.to_dict()})
    adata.uns.update({"ligands_pval": ligand_pval.to_dict()})
    adata.uns.update({"receptors_pval": receptor_pval.to_dict()})
    adata.uns.update({"ligands_corr_pval": ligand_corr_pval.to_dict()})
    adata.uns.update({"receptors_corr_pval": receptor_corr_pval.to_dict()})
    adata.uns.update({"specificity_setting": settings})

    if return_values:
        return adata, ligand_values, receptor_values

    return adata

# Save and load specificity calculations (time consuming)

def save_specificity(adata, filename):
    """Saves data calculated by cn.connect.specificity to an excel file. 
    This file can later be loaded using cn.connect.load_specificity"""
    import pandas as pd
    
    keys = [
        'ligands',
        'receptors',
        'ligands_zscore',
        'receptors_zscore',
        'ligands_pval',
        'receptors_pval',
        'ligands_pval',
        'receptors_pval',
        'ligands_corr_pval',
        'receptors_corr_pval']
    
    xls = pd.ExcelWriter(filename)
    for key in keys:
        table = pd.DataFrame(adata.uns[key])
        table.to_excel(xls, sheet_name=key)
    s = pd.Series(adata.uns["specificity_setting"])
    s.to_excel(xls, sheet_name="specificity_setting")
    xls.close()

def load_specificity(adata, filename):
    """Loads previously calculated specificity to an andata object"""

    import pandas as pd
    keys = [
        'ligands',
        'receptors',
        'ligands_zscore',
        'receptors_zscore',
        'ligands_pval',
        'receptors_pval',
        'ligands_pval',
        'receptors_pval',
        'ligands_corr_pval',
        'receptors_corr_pval']

    for key in keys:
        data = pd.read_excel(filename, sheet_name=key, index_col=0)
        adata.uns[key] = {k:value.to_dict() for k, value in data.iteritems()}
    adata.uns["specificity_setting"] = pd.read_excel(filename, sheet_name="specificity_setting", index_col=0).to_dict()[0]
    


# Used to add molecules manually. Check that ligands are named
# correcly as in the Ligands table
# Current version expect mouse genes, and automatically find orthogonal 
# genes when a species specific library is constructed. These files can be edited manually if needed


def get_molecules():
    molecules = [{
        "ligand": "L-glutamic acid",
        "type": "molecule",
        "synthesis": None,
        "transport": ["Slc17a6", "Slc17a7", "Slc17a8"],
        "reuptake": ["Slc1a1", "Slc1a2", "Slc1a6"],
        "excluded": None
    }, {
        "ligand": "GABA",
        "type": "molecule",
        "synthesis": ["Gad1", "Gad2"],
        "transport": ["Slc32a1"],
        "reuptake": ["Slc6a1"],
        "excluded": None
    }, {
        "ligand": "glycine",
        "type": "molecule",
        "synthesis": None,
        "transport": ["Slc32a1"],
        "reuptake": ["Slc6a9"],
        "excluded": None
    }, {
        "ligand": "dopamine",
        "type": "molecule",
        "synthesis": ["Th", "Ddc"],
        "transport": ["Slc18a2", "Slc18a3"],
        "reuptake": ["Slc6a3"],
        "excluded": ["Dbh"]
    }, {
        "ligand": "(-)-noradrenaline",
        "type": "molecule",
        "synthesis": ["Th", "Ddc", "Dbh"],
        "transport": ["Slc18a2", "Slc18a3"],
        "reuptake": ["Slc6a3"],
        "excluded": ["Pnmt"]
    }, {
        "ligand": "(-)-adrenaline",
        "type": "molecule",
        "synthesis": ["Th", "Ddc", "Dbh", "Pnmt"],
        "transport": ["Slc18a2"],
        "reuptake": None,
        "excluded": None
    }, {
        "ligand": "5-hydroxytryptamine",
        "type": "molecule",
        "synthesis": ["Tph2", "Ddc"],
        "transport": ["Slc18a2"],
        "reuptake": ["Slc6a4"],
        "excluded": None
    }, {
        "ligand": "acetylcholine",
        "type": "molecule",
        "synthesis": ["Chat"],
        "transport": ["Slc18a3"],
        "reuptake": ["Ache", "Slc5a7"],
        "excluded": None
    }]
    return molecules
